import pytest

import numpy as np
from astropy.io import fits

from anisocado import misc

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm


@pytest.fixture(scope="class")
def basic_fv_psf():
    n = 30
    coords = [(-n, n),  (0, n),  (n, n),
              (-n, 0),  (0, 0),  (n, 0),
              (-n, -n), (0, -n), (n, -n)]
    waves = [0.9, 1.1, 1.6, 2.15]
    hdu = misc.make_simcado_psf_file(coords=coords, wavelengths=waves, N=128)

    return hdu


class TestSimcadoPsfFile:
    def test_throws_error_with_no_input(self):
        with pytest.raises(TypeError):
            misc.make_simcado_psf_file()

    def test_returns_hdulist_for_basic_input(self):
        hdu = misc.make_simcado_psf_file(coords=[(0, 0)], wavelengths=[2.15],
                                                   N=128)
        assert isinstance(hdu, fits.HDUList)
        assert isinstance(hdu[1], fits.BinTableHDU)
        assert len(hdu) == 3
        assert hdu[2].data.shape == (1, 128, 128)

    def test_returns_full_hdulist_for_full_input(self):
        n = 30
        coords = [(0, 0), (-n, -n), (n, -n), (n, n), (-n, n)]
        waves = [0.9, 1.1, 1.6, 2.15]
        hdu = misc.make_simcado_psf_file(coords=coords, wavelengths=waves,
                                         N=128)
        assert len(hdu) == len(waves) + 2
        assert hdu[2].data.shape == (len(coords), 128, 128)
        assert hdu[2].header["CDELT1"] == 0.004 / 3600.


@pytest.mark.usefixtures("basic_fv_psf")
class TestFvpsfFileConsistencyChecks:
    def test_has_correct_keywords_in_ext0(self, basic_fv_psf):
        for key in ["AUTHOR", "DATE_CRE", "DATE_MOD", "SOURCE", "STATUS",
                    "ETYPE", "ECAT", "EDATA"]:
            assert key in basic_fv_psf[0].header

    def test_has_catalogue_psf(self, basic_fv_psf):
        ecat = basic_fv_psf[0].header["ECAT"]
        cat_hdu = basic_fv_psf[ecat]
        for key in ["NUMPSFS", "CATTYPE", "CUNIT1"]:
            assert key in cat_hdu.header

        if cat_hdu.header["CATTYPE"] == "table":
            assert isinstance(cat_hdu.data, (fits.BinTableHDU, fits.FITS_rec))

        elif basic_fv_psf[ecat].header["CATTYPE"] == "image":
            assert isinstance(cat_hdu.data, fits.ImageHDU)
            for key in ["CRVAL1", "CRPIX1", "CDELT1"]:
                assert key in cat_hdu.header

    def test_has_the_right_number_of_layers_per_psf_hdu(self, basic_fv_psf):
        ecat = basic_fv_psf[0].header["ECAT"]
        cat_hdu = basic_fv_psf[ecat]
        n_layers = len(set(cat_hdu.data["layer"]))

        edata = basic_fv_psf[0].header["EDATA"]
        data_indexes = range(edata, len(basic_fv_psf))
        for ii in data_indexes:
            assert basic_fv_psf[ii].data.shape[0] == n_layers
            for key in ["CRVAL1", "CRPIX1", "CDELT1", "WAVE0"]:
                assert key in basic_fv_psf[ii].header

    def plot_psfs(self, basic_fv_psf):
        for psf_hdu in basic_fv_psf[2:]:
            for i in range(len(psf_hdu.data)):
                plt.subplot(3, 3, i + 1)
                plt.imshow(psf_hdu.data[i, :, :].T, origin="l", norm=LogNorm())

            plt.show()

    def print_coords(self):
        #coords = psf.field_positions_for_simcado_psf()
        #waves = [0.8, 0.9, 1.05, 1.2, 1.5, 1.7, 2.0, 2.2]
        #hdu = psf.make_simcado_psf_file()

        map = misc.make_strehl_map_from_table([(0, 0)])
        plt.imshow(map)
        plt.show()
        print(np.unique(map))




import pytest

import numpy as np

from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

from anisocado import AnalyticalScaoPsf
from anisocado.misc import on_axis_strehl_for_kernel_size

PLOTS = False


class TestMakeStrehlMap:
    def test_make_a_strehl_map_over_a_grid_on_the_focal_plane(self):
        psf = AnalyticalScaoPsf(N=128, wavelength=2.15)
        x, y = np.mgrid[-50:51:20, -50:51:20]

        sr = np.zeros(x.shape)
        for i in range(len(x)):
            for j in range(len(y)):
                psf.shift_off_axis(x[i, j], y[i, j])
                sr[i, j] = psf.strehl_ratio
        print(np.max(sr))

        if PLOTS:
            plt.imshow(sr)  # norm=LogNorm())
            plt.contourf(x, y, sr, np.arange(0, 1, 0.05))
            plt.colorbar()
            plt.show()


class TestStrehlRatiosForDifferentSizedKernels:
    def test_how_much_does_the_SR_change_for_kernel_size_increase(self):
        side_length = [64, 128, 256, 512, 1024]
        sr = on_axis_strehl_for_kernel_size(side_length, wavelength=2.15)
        if PLOTS:
            plt.plot(side_length, sr)
            plt.show()


class TestPsfsAlongXaxis:
    def test_strehl_ratio_along_x_axis(self):
        for wave in [0.8, 0.9, 1.1, 1.6, 2.15]:
            psf = AnalyticalScaoPsf(N=128, wavelength=wave)
            x, y = np.mgrid[0:41:1, 0:1:1]

            sr = np.zeros(x.shape)
            for i in range(x.shape[0]):
                for j in range(x.shape[1]):
                    psf.shift_off_axis(x[i, j], y[i, j])
                    sr[i, j] = psf.strehl_ratio

            if PLOTS:
                print(wave, np.sum(psf._kernel_sum))
                plt.plot(x.flatten(), sr, label="{} um".format(wave))

        if PLOTS:
            # plt.semilogy()
            plt.legend()
            plt.xlabel("Distance from NGS [arcsec]")
            plt.ylabel("Strehl Ratio")
            plt.show()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  8 15:38:32 2018

@author: Eric Gendron

Original Title : make_ricoPupil.py

"""

import numpy as np
# import matplotlib.pyplot as plt
# plt.ion()

"""
CONVENTIONS : Ce fichier est ecrit en convention X,Y

...................................................... 1.0 Convention X,Y 

Dans la convention d'axes (x,y), tous les tableaux representant des images
s'adressent par
    tab[ix, iy]
Les indices s'utilisent alors classiquement
    ind = np.where(tab2D > 0)
    (ix, iy) = np.where(tab2D > 0)
où ind[0] porte X. On fera donc appel a
    tab[ind]
    tab[ind[0], tab[1]]
    tab[ix, iy]

La fonction meshgrid fait chier, et doit s'appeler par
    X,Y = meshgrid(x,y, indexing='ij')
pour creer des tableaux conformes a un appel en X[ix, iy].

Dans les tableaux qui stockent des coord en x,y on utilise
x = tab[:,0]
y = tab[:,1]
pour que l'utilisation de flatten() mette X en premier. Cette facon de proceder
est logique/compatible avec les indices.

Dans les fonctions on place x d'abord, y ensuite, dans les arguments comme
dans le retour de fonction
    def toto(...,x,y,...)
        return x, y
et
    x, y = toto(...)

Graphisme: Pour afficher un tel tableau et avoir une representation
'naturelle' avec x "a droite" et y "en haut" il faut par contre
utiliser une transposition et retournement d'axe
    plt.imshow(tab.T, origin='lower')
et qui offre une utilisation classique de
    plt.plot(x, y, ...)
qui placera un overlay d'un plot de façon coherente sur l'image affichee.


Une sortie FITS d'un tableau tab[x,y] va generer un fichier qui contient des
data avec un axe rapide NAXIS1 dirige selon Y. Attention car de nombreux
logiciels considerent axe rapide = axe x (par exemple ds9), ou representent 
l'axe rapide souvent horizontal en natif (ds9, python, yorick, idl, ...)
Donc, pour les entrees/sorties on transposera les data
    pf.writeto('monfichier.fits', tab.T)
et
    tab = pf.getdata('monfichier.fits').T


...................................................... 2.0 Convention Y,X 
Dans la convention d'axe (y,x), tous les tableaux d'images sont
adresses par
    tab[iy, ix]

Pour une manipulation d'indices du genre
    ind = np.where(tab2D > 0)
on utilise l'une des 3 possibilites suivantes
    ix = ind[1]; iy = ind[0]       # attention a bien swapper y=0/x=1
    (iy, ix) = ind                 # y vient d'abord, coherent avec les tableaux
    (iy, ix) = where(tab2D > 0)    # y vient d'abord
et appeler les elements des tableaux par
    tab[iy, ix]
    tab[ind]

Toutes les autres notations restent en "x d'abord, y ensuite".

Dans une procedure de calcul de coordonnees, on gardera la convention
    def toto(args, x, y, ):
        ...
        return x, y
et a l'appel de la fonction et recup des coordonnees on garde
    x, y = toto(args, ax, ay, ..)

La fonction meshgrid doit s'appeler par
    X,Y = meshgrid(x,y)
pour creer des tableaux conformes a un appel en X[iy, ix].

Dans les tableaux qui stockent des coord en x,y on utilise
x = tab[:,0]
y = tab[:,1]
pour que l'utilisation de flatten() mette X en premier. Cependant cette
notation est en conflit avec le traitement des indices (fonction where()) qui
placent Y d'abord.

Graphisme: Pour afficher un tel tableau et avoir une representation
'naturelle' avec x "a droite" et y "en haut" il faut juste utiliser le
retournement d'axe
    plt.imshow(tab, origin='lower')
suivi d'une utilisation classique de
    plt.plot(x, y, ...)
qui placera un overlay d'un plot de façon coherente sur l'image affichee.


Une sortie FITS d'un tableau tab[x,y] cree un fichier qui aura un axe
rapide NAXIS1 selon X, coherent avec la plupart des logiciels (ds9, python,
yorick, idl, ...)


............................................................ 3.0 Meshgrid

Quand on definit
x = ...
y = ...
et qu'on utilise np.meshgrid(), on a 

   code                            convention d'appel        mat/imshow      
X,Y = meshgrid(x,y)                     [y,x]               X horizontal
X,Y = meshgrid(x,y,indexing='ij')       [x,y]               X vertical  :-(
Y,X = meshgrid(y,x)                     [x,y]               X vertical  :-(
Y,X = meshgrid(y,x,indexing='ij')       [y,x]               X horizontal
X,Y = meshgrid(y,x)                   incoherent

meshgrid(x,y) equivaut a indexing='xy' (default).
Indexing 'ij' ou 'xy' affecte la transposition des X,Y de sortie.
L'interversion x/y dans les parametres d'appel et de sortie affecte la
transposition des X/Y.


Donc 
- soit on utilise meshgrid(x,y) sans option d'indexing, ce qui donne des
plot 2D plutot user-friendly sans pencher la tete et sans transposee, mais
la notation d'appel dans les tableaux doit etre [iy,ix]
- soit on utilise meshgrid(..,indexing='ij'), on aura avec imshow() sans
option des plots a regarder avec la tete 90° a droite, et une notation [x,y]
partout dans le code.
Pour avoir les plots python "dans le bon sens" avec une notation [x,y] il faut
plt.imshow(p.T,origin='lower').


L'utilisation
    X,Y = meshgrid(y,x)
ou
    Y,X = meshgrid(x,y)
n'a de sens que si x==y (par exemple x=y=np.linspace(-1,1,n) ...) et permet
le meme rendu que indexing='ij'. Mais c'est un joli hasard. Des que les
axes x et y se distinguent, soit par le nbre de points, soit par les
valeurs (soit le range, le step, l'offset, etc.) alors inverser x et y est
juste purement incoherent.


"""


def fillPolygon(x, y, i0, j0, scale, gap, N, index=0):
    """
    From a list of points defined by their 2 coordinates list
    x and y, creates a filled polygon with sides joining the points.
    The polygon is created in an image of size (N, N).
    The origin (x,y)=(0,0) is mapped at pixel i0, j0 (both can be
    floating-point values).
    Arrays x and y are supposed to be in unit U, and scale is the
    pixel size in U units.

    :returns: filled polygon (N, N), boolean
    :param float x, y: list of points defining the polygon
    :param float i0, j0: index of pixels where the pupil should be centred.
                         Can be floating-point indexes.
    :param float scale: size of a pixel of the image, in same unit as x and y.
    :param float N: size of output image.

    :Example:
    x = np.array([1,-1,-1.5,0,1.1])
    y = np.array([1,1.5,-0.2,-2,0])
    N = 200
    i0 = N/2
    j0 = N/2
    gap = 0.
    scale = 0.03
    pol = fillPolygon(x, y, i0, j0, scale, gap, N, index=2)

    """
    # define coordinates map centred on (i0,j0) with same units as x,y.
    X = (np.arange(N) - i0) * scale
    Y = (np.arange(N) - j0) * scale
    X, Y = np.meshgrid(X, Y, indexing='ij')  # indexage [x,y]

    # define centre of polygon x0, y0
    x0 = np.mean(x)
    y0 = np.mean(y)

    # compute angles of all pixels coordinates of the map, and all
    # corners of the polygon
    T = (np.arctan2(Y - y0, X - x0) + 2 * np.pi) % (2 * np.pi)
    t = (np.arctan2(y - y0, x - x0) + 2 * np.pi) % (2 * np.pi)

    # on va voir dans quel sens ca tourne. Je rajoute ca pour que ca marche
    # quel que soit le sens de rotation des points du polygone.
    # En fait, j'aurais peut etre pu classer les points par leur angle, pour
    # etre sur que ca marche meme si les points sont donnes dans ts les cas
    sens = np.median(np.diff(t))
    if sens < 0:
        x = x[::-1]
        y = y[::-1]
        t = t[::-1]

    # re-organise order of polygon points so that it starts from
    # angle = 0, or at least closest to 0.
    imin = t.argmin()  # position of the minimum
    if imin != 0:
        x = np.roll(x, -imin)
        y = np.roll(y, -imin)
        t = np.roll(t, -imin)

    # For each couple of consecutive corners A, B, of the polygon, one fills
    # the triangle AOB with True.
    # Last triangle has a special treatment because it crosses the axis
    # with theta=0=2pi
    n = x.shape[0]  # number of corners of polygon
    indx, indy = (np.array([], dtype=np.int), np.array([], dtype=np.int))
    distedge = np.array([], dtype=np.float)
    for i in range(n):
        j = i + 1  # j=element next i except when i==n : then j=0 (cycling)
        if j == n:
            j = 0
            sub = np.where((T >= t[-1]) | (T <= (t[0])))
        else:
            sub = np.where((T >= t[i]) & (T <= t[j]))
        # compute unitary vector des 2 sommets
        dy = y[j] - y[i]
        dx = x[j] - x[i]
        vnorm = np.sqrt(dx ** 2 + dy ** 2)
        dx /= vnorm
        dy /= vnorm
        # calcul du produit vectoriel
        crossprod = dx * (Y[sub] - y[i]) - dy * (X[sub] - x[i])
        tmp = crossprod > gap
        indx = np.append(indx, sub[0][tmp])
        indy = np.append(indy, sub[1][tmp])
        distedge = np.append(distedge, crossprod[tmp])

    # choice of what is returned : either only the indexes, or the
    # boolean map
    if index == 1:
        return (indx, indy, distedge)
    elif index == 2:
        a = np.zeros((N, N))
        a[indx, indy] = distedge
        return a
    else:
        a = np.zeros((N, N), dtype=np.bool)
        a[indx, indy] = True  # convention [x,y]

    return a


def centrePourVidal(N, i0, j0, centerMark):
    """
    Renvoie une image de boolens (False) de taille (N,N) avec un point
    ou une croix (True) centree sur (i0, j0).
    :param int N: taille de l'image de sortie
    :param float i0, j0: position du marqueur de sortie
    :param int centerMark: 0 (pour rien), 1 (option point) ou 2 (option croix)
    """
    scale = 1.0
    res = 0
    X = (np.arange(N) - i0) * scale
    Y = (np.arange(N) - j0) * scale
    X, Y = np.meshgrid(X, Y, indexing='ij')  # convention d'appel [x,y]
    if centerMark == 1:
        res = (X ** 2 + Y ** 2) < 1
    if centerMark == 2:
        res = (np.abs(X) < 0.9) | (np.abs(Y) < 0.9)
    return res


def fillSpider(N, nspider, dspider, i0, j0, scale, rot):
    """
    Creates a boolean spider mask on a map of dimensions (N,N)
    The spider is centred at floating-point coords (i0,j0).

    :returns: spider image (boolean)
    :param int N: size of output image
    :param int nspider: number of spiders
    :param float dspider: width of spiders
    :param float i0: coord of spiders symmetry centre
    :param float j0: coord of spiders symmetry centre
    :param float scale: size of a pixel in same unit as dspider
    :param float rot: rotation angle in radians

    """
    a = np.ones((N, N), dtype=np.bool)
    X = (np.arange(N) - i0) * scale
    Y = (np.arange(N) - j0) * scale
    X, Y = np.meshgrid(X, Y, indexing='ij')  # convention d'appel [x,y]
    w = 2 * np.pi / nspider
    # rot += np.pi/2  # parce que c'est comme ca !!
    for i in range(nspider):
        nn = (abs(
            X * np.cos(i * w - rot) + Y * np.sin(i * w - rot)) < dspider / 2.)
        a[nn] = False
    return a


def generateEeltPupil_slow(npt, dspider, i0, j0, pixscale, rotdegree):
    """
    Computes the binary EELT pupil on a map of size (npt, npt).
    This is the original function, that builds the pupil shape according to
    hardcoded contours.
    This function is now obsolete, because it's been replaced by the faster
    one generateEeltPupilMask()

    :returns: pupil image (npt, npt), boolean
    :param float dspider: width of spiders in meters
    :param float i0, j0: index of pixels where the pupil should be centred.
                         Can be floating-point indexes.
    :param float pixscale: size of a pixel of the image, in meters.
    :param float rotdegree: rotation angle of the pupil, in degrees.

    :Example:
    >>> pup = generateEeltPupil_slow(800, 0.6, 400, 400, 0.1, 3.0)

    """
    x = np.array(
        [18.4524, 18.798, 18.4514, 18.796, 18.4484, 18.7919, 18.4433, 18.7858,
         18.4363, 18.7776,
         18.4273, 17.7349, 17.3831, 17.7243, 17.3717, 16.6772, 16.3233, 16.6645,
         16.3099,
         15.6135, 15.2579, 15.5991, 15.2429, 14.545, 14.188, 14.5292, 14.1718,
         13.4727, 13.1146,
         12.4138, 12.0552, 11.3528, 10.9939, 10.2902, 9.93103, 9.22619, 8.86699,
         8.16118, 7.8021,
         7.09552, 6.73671, 6.02955, 5.67117, 4.96362, 4.60582, 3.89808, 3.54699,
         2.83805,
         2.48141, 1.77263, 1.41934, 0.709707, 0.354564, -0.354564, -0.709707,
         -1.41934,
         -1.77263, -2.48141, -2.83805, -3.54699, -3.89808, -4.60582, -4.96362,
         -5.67117,
         -6.02955, -6.73671, -7.09552, -7.8021, -8.16118, -8.86699, -9.22619,
         -9.93103,
         -10.2902, -10.9939, -11.3528, -12.0552, -12.4138, -13.1146, -13.4727,
         -14.1718,
         -14.5292, -14.188, -14.545, -15.2429, -15.5991, -15.2579, -15.6135,
         -16.3099, -16.6645,
         -16.3233, -16.6772, -17.3717, -17.7243, -17.3831, -17.7349, -18.4273,
         -18.7776,
         -18.4363, -18.7858, -18.4433, -18.7919, -18.4484, -18.796, -18.4514,
         -18.798, -18.4524,
         -18.798, -18.4514, -18.796, -18.4484, -18.7919, -18.4433, -18.7858,
         -18.4363, -18.7776,
         -18.4273, -17.7349, -17.3831, -17.7243, -17.3717, -16.6772, -16.3233,
         -16.6645,
         -16.3099, -15.6135, -15.2579, -15.5991, -15.2429, -14.545, -14.188,
         -14.5292, -14.1718,
         -13.4727, -13.1146, -12.4138, -12.0552, -11.3528, -10.9939, -10.2902,
         -9.93103,
         -9.22619, -8.86699, -8.16118, -7.8021, -7.09552, -6.73671, -6.02955,
         -5.67117,
         -4.96362, -4.60582, -3.89808, -3.54699, -2.83805, -2.48141, -1.77263,
         -1.41934,
         -0.709707, -0.354564, 0.354564, 0.709707, 1.41934, 1.77263, 2.48141,
         2.83805, 3.54699,
         3.89808, 4.60582, 4.96362, 5.67117, 6.02955, 6.73671, 7.09552, 7.8021,
         8.16118, 8.86699,
         9.22619, 9.93103, 10.2902, 10.9939, 11.3528, 12.0552, 12.4138, 13.1146,
         13.4727,
         14.1718, 14.5292, 14.188, 14.545, 15.2429, 15.5991, 15.2579, 15.6135,
         16.3099, 16.6645,
         16.3233, 16.6772, 17.3717, 17.7243, 17.3831, 17.7349, 18.4273, 18.7776,
         18.4363,
         18.7858, 18.4433, 18.7919, 18.4484, 18.796, 18.4514, 18.798])

    y = np.array(
        [0, 0.614323, 1.22918, 1.84277, 2.45796, 3.07061, 3.68594, 4.29746,
         4.91271, 5.5229,
         6.13789, 6.14356, 6.75902, 7.36787, 7.98272, 7.98968, 8.60474, 9.21184,
         9.82596,
         9.83398, 10.448, 11.053, 11.6658, 11.6747, 12.2871, 12.8896, 13.5004,
         13.51, 14.1202,
         14.1294, 14.7389, 14.7478, 15.3564, 15.3648, 15.9724, 15.9802, 16.5867,
         16.5939,
         17.1992, 17.2057, 17.8095, 17.8154, 18.4177, 18.4227, 19.0233, 19.0275,
         18.4307,
         18.4337, 19.0337, 19.0357, 18.4377, 18.4387, 19.0378, 19.0378, 18.4387,
         18.4377,
         19.0357, 19.0337, 18.4337, 18.4307, 19.0275, 19.0233, 18.4227, 18.4177,
         17.8154,
         17.8095, 17.2057, 17.1992, 16.5939, 16.5867, 15.9802, 15.9724, 15.3648,
         15.3564,
         14.7478, 14.7389, 14.1294, 14.1202, 13.51, 13.5004, 12.8896, 12.2871,
         11.6747, 11.6658,
         11.053, 10.448, 9.83398, 9.82596, 9.21184, 8.60474, 7.98968, 7.98272,
         7.36787, 6.75902,
         6.14356, 6.13789, 5.5229, 4.91271, 4.29746, 3.68594, 3.07061, 2.45796,
         1.84277, 1.22918,
         0.614323, 0, -0.614323, -1.22918, -1.84277, -2.45796, -3.07061,
         -3.68594,
         -4.29746, -4.91271, -5.5229, -6.13789, -6.14356, -6.75902, -7.36787,
         -7.98272,
         -7.98968, -8.60474, -9.21184, -9.82596, -9.83398, -10.448, -11.053,
         -11.6658, -11.6747,
         -12.2871, -12.8896, -13.5004, -13.51, -14.1202, -14.1294, -14.7389,
         -14.7478, -15.3564,
         -15.3648, -15.9724, -15.9802, -16.5867, -16.5939, -17.1992, -17.2057,
         -17.8095,
         -17.8154, -18.4177, -18.4227, -19.0233, -19.0275, -18.4307, -18.4337,
         -19.0337,
         -19.0357, -18.4377, -18.4387, -19.0378, -19.0378, -18.4387, -18.4377,
         -19.0357,
         -19.0337, -18.4337, -18.4307, -19.0275, -19.0233, -18.4227, -18.4177,
         -17.8154,
         -17.8095, -17.2057, -17.1992, -16.5939, -16.5867, -15.9802, -15.9724,
         -15.3648,
         -15.3564, -14.7478, -14.7389, -14.1294, -14.1202, -13.51, -13.5004,
         -12.8896, -12.2871,
         -11.6747, -11.6658, -11.053, -10.448, -9.83398, -9.82596, -9.21184,
         -8.60474, -7.98968,
         -7.98272, -7.36787, -6.75902, -6.14356, -6.13789, -5.5229, -4.91271,
         -4.29746,
         -3.68594, -3.07061, -2.45796, -1.84277, -1.22918, -0.614323])

    # Rotation matrices
    rot = rotdegree * np.pi / 180.00
    mrot = np.array([[np.cos(rot), np.sin(rot)], [-np.sin(rot), np.cos(rot)]])

    # rotation of coordinates of outer segments
    u = mrot.dot([x, y])
    pup = fillPolygon(u[0], u[1], i0, j0, pixscale, 0., npt)

    # INTERNAL CONTOUR OF OBSCURATION
    xx = np.array(
        [5.02571, 4.66726, 5.02543, 4.66673, 5.02458, 4.66568, 3.94877, 3.58959,
         2.87216,
         2.51286, 1.7951, 1.43584, 0.717959, 0.358899, -0.358899, -0.717959,
         -1.43584, -1.7951,
         -2.51286, -2.87216, -3.58959, -3.94877, -4.66568, -5.02458, -4.66673,
         -5.02543,
         -4.66726, -5.02571, -4.66726, -5.02543, -4.66673, -5.02458, -4.66568,
         -3.94877,
         -3.58959, -2.87216, -2.51286, -1.7951, -1.43584, -0.717959, -0.358899,
         0.358899,
         0.717959, 1.43584, 1.7951, 2.51286, 2.87216, 3.58959, 3.94877, 4.66568,
         5.02458,
         4.66673, 5.02543, 4.66726])

    yy = np.array(
        [0., 0.62184, 1.24347, 1.86531, 2.48652, 3.10816, 3.10885, 3.73041,
         3.73104,
         4.35239, 4.35288, 4.97389, 4.97416, 5.59468, 5.59468, 4.97416, 4.97389,
         4.35288,
         4.35239, 3.73104, 3.73041, 3.10885, 3.10816, 2.48652, 1.86531, 1.24347,
         0.62184,
         0.0, -0.62184, -1.24347, -1.86531, -2.48652, -3.10816, -3.10885,
         -3.73041,
         -3.73104, -4.35239, -4.35288, -4.97389, -4.97416, -5.59468, -5.59468,
         -4.97416,
         -4.97389, -4.35288, -4.35239, -3.73104, -3.73041, -3.10885, -3.10816,
         -2.48652,
         -1.86531, -1.24347, -0.62184])

    # rotation of coordinates of inner segments (central obs)
    u = mrot.dot([xx, yy])
    pup = pup & ~fillPolygon(u[0], u[1], i0, j0, pixscale, 0, npt)

    # SPIDERS ............................................
    nspider = 3  # pour le jour ou on voudra plus de spiders..
    if (dspider > 0 and nspider > 0):
        pup = pup & fillSpider(npt, nspider, dspider, i0, j0, pixscale, rot)

    return pup


def createHexaPattern(pitch, supportSize):
    """
    Cree une liste de coordonnees qui decrit un maillage hexagonal.
    Retourne un tuple (x,y).

    Le maillage est centre sur 0, l'un des points est (0,0).
    Une des pointes de l'hexagone est dirigee selon l'axe Y, au sens ou le
    tuple de sortie est (x,y).

    :param float pitch: distance between 2 neighbour points
    :param int supportSize: size of the support that need to be populated

    """
    V3 = np.sqrt(3)
    nx = int(np.ceil((supportSize / 2.0) / pitch) + 1)
    x = pitch * (np.arange(2 * nx + 1) - nx)
    ny = int(np.ceil((supportSize / 2.0) / pitch / V3) + 1)
    y = (V3 * pitch) * (np.arange(2 * ny + 1) - ny)
    x, y = np.meshgrid(x, y, indexing='ij')
    x = x.flatten()
    y = y.flatten()
    peak_axis = np.append(x, x + pitch / 2.)  # axe dirige selon sommet
    flat_axis = np.append(y, y + pitch * V3 / 2.)  # axe dirige selon plat
    return flat_axis, peak_axis


def generateCoordSegments(D, rot):
    """
    Computes the coordinates of the corners of all the hexagonal
    segments of M1.
    Result is a tuple of arrays(6, 798).

    :param float D: D is the pupil diameter in meters, it must be set to 40.0 m
    for the nominal EELT.
    :param float rot: pupil rotation angle in radians

    """
    V3 = np.sqrt(3)
    pitch = 1.227314  # no correction du bol
    pitch = 1.244683637214  # diametre du cerle INSCRIT
    # diamseg = pitch*2/V3  # diametre du cercle contenant TOUT le segment
    # print("segment diameter : %.6f\n" % diamseg)

    # Creation d'un pattern hexa avec pointes selon la variable <ly>
    lx, ly = createHexaPattern(pitch, 35 * pitch)
    ll = np.sqrt(lx ** 2 + ly ** 2)
    # Elimination des segments non valides grace a 2 nombres parfaitement
    # empiriques ajustes a-la-mano.
    inner_rad, outer_rad = 4.1, 15.4  # nominal, 798 segments
    nn = (ll > inner_rad * pitch) & (ll < outer_rad * pitch);
    lx = lx[nn]
    ly = ly[nn]
    lx, ly = reorganizeSegmentsOrderESO(lx, ly)
    ll = np.sqrt(lx ** 2 + ly ** 2)

    # n = ll.shape[0]
    # print("Nbre de segments : %d\n" % n)
    # Creation d'un hexagone-segment avec pointe dirigee vers
    # variable <hx> (d'ou le cos() sur hx)
    th = np.linspace(0, 2 * np.pi, 7)[0:6]
    hx = np.cos(th) * pitch / V3
    hy = np.sin(th) * pitch / V3

    # Le maillage qui permet d'empiler des hexagones avec sommets 3h-9h
    # est un maillage hexagonal avec sommets 12h-6h, donc a 90°.
    # C'est pour ca qu'il a fallu croiser les choses avant.
    x = (lx[None, :] + hx[:, None])
    y = (ly[None, :] + hy[:, None])
    r = np.sqrt(x ** 2 + y ** 2)
    R = 95.7853
    rrc = R / r * np.arctan(r / R)  # correction factor
    x *= rrc
    y *= rrc

    nominalD = 40.0  # size of the OFFICIAL E-ELT
    if D != nominalD:
        x *= D / nominalD
        y *= D / nominalD

    # Rotation matrices
    mrot = np.array([[np.cos(rot), np.sin(rot)], [-np.sin(rot), np.cos(rot)]])

    # rotation of coordinates
    # le tableau [x,y] est de taille (2,6,798). Faut un transpose a la con
    # pour le transformer en (6,2,798) pour pouvoir faire le np.dot
    # correctement. En sortie, xrot est (2,6,798).
    xyrot = np.dot(mrot, np.transpose(np.array([x, y]), (1, 0, 2)))

    return xyrot[0], xyrot[1]


def reorganizeSegmentsOrderESO(x, y):
    """
    Reorganisation des segments facon ESO.
    Voir
    ESO-193058 Standard Coordinate System and Basic Conventions

    :param float x: tableau des centres X des segments
    :param float y: idem Y
    :return tuple (x,y): meme tuple que les arguments d'entree, mais tries.

    """
    # pi/2, pi/6, 2.pi, ...
    pi_3 = np.pi / 3
    pi_6 = np.pi / 6
    pix2 = 2 * np.pi
    # calcul des angles
    t = (np.arctan2(y, x) + pi_6 - 1e-3) % (pix2)
    X = np.array([])
    Y = np.array([])
    A = 100.
    for k in range(6):
        sector = (t > k * pi_3) & (t < (k + 1) * pi_3)
        u = k * pi_3
        distance = (A * np.cos(u) - np.sin(u)) * x[sector] + (
                    np.cos(u) + A * np.sin(u)) * y[sector]
        indsort = np.argsort(distance)
        X = np.append(X, x[sector][indsort])
        Y = np.append(Y, y[sector][indsort])
    return X, Y


def getdatatype(truc):
    """
    Returns the data type of a numpy variable, either scalar value or array.

    """
    if np.isscalar(truc):
        return type(truc)
    else:
        return type(truc.flatten()[0])


def generateSegmentProperties(attribute, hx, hy, i0, j0, scale, gap, N, D,
                              softGap=0):
    """
    Builds a 2D image of the pupil with some attributes for each of the
    segments. Those segments are described from arguments hx and hy, that
    are produced by the function generateCoordSegments(D, rot).

    When attribute is a phase, then it must be a float array of dimension
    [3, 798] with the dimension 3 being piston, tip, and tilt.
    Units of phase is xxx rms, and the output of the procedure will be
    in units of xxx.


    :returns: pupil image (N, N), with the same type of input argument attribute

    :param float/int/bool attribute: scalar value or 1D-array of the
        reflectivity of the segments or 2D array of phase
           If attribute is scalar, the value will be replicated for all segments
           If attribute is a 1D array, then it shall contain the reflectivities
           of all segments.
           If attribute is a 2D array then it shall contain the piston, tip
           and tilt of the segments. The array shall be of dimension
           [3, 798] that contains [piston, tip, tilt]
           On output, the data type of the pupil map will be the same as input
    :param float hx, hy: arrays [6,:] describing the segment shapes. They are
        generated using generateCoordSegments()
    :param float dspider: width of spiders in meters
    :param float i0, j0: index of pixels where the pupil should be centred.
                         Can be floating-point indexes.
    :param float scale: size of a pixel of the image, in meters.
    :param float gap: half-space between segments in meters
    :param int N: size of the output array (N,N)
    :param float D: diameter of the pupil. For the nominal EELT, D shall
                    be set to 40.0
    :param bool softGap: if False, the gap between segments is binary 0/1
          depending if the pixel is within the gap or not. If True, the gap
          is a smooth region of a fwhm of 2 pixels with a depth related to the
          gap width.



    """

    # number of segments
    nseg = hx.shape[-1]
    # If <attribute> is a scalar, then we make a list. It will be required
    # later on to set the attribute to each segment.
    if np.isscalar(attribute):
        attribute = np.array([attribute] * nseg)

    # the pupil map is created with the same data type as <attribute>
    pupil = np.zeros((N, N), dtype=getdatatype(attribute))

    # average coord of segments
    x0 = np.mean(hx, axis=0)
    y0 = np.mean(hy, axis=0)
    # avg coord of segments in pixel indexes
    x0 = x0 / scale + i0
    y0 = y0 / scale + j0
    # size of mini-support
    hexrad = 0.75 * D / 40. / scale
    ix0 = np.floor(x0 - hexrad).astype(int) - 1
    iy0 = np.floor(y0 - hexrad).astype(int) - 1
    segdiam = np.ceil(hexrad * 2 + 1).astype(int) + 1

    n = attribute.shape[0]
    if n != 3:
        # attribute is a signel value : either reflectivity, or boolean,
        # or just piston.
        if softGap != 0:
            # Soft gaps
            # The impact of gaps are modelled using a simple function:
            # Lorentz, 1/(1+x**2)
            # The fwhm is always equal to 2 pixels because the gap is supposed
            # to be "small/invisible/undersampled". The only visible thing is
            # the width of the impulse response, chosen 2-pixel wide to be
            # well sampled.
            # The "depth" is related to the gap width. The integral of a
            # Lorentzian of 2 pix wide is PI. Integral of a gap of width 'gap'
            # in pixels is 'gap'.
            # So the depth equals to gap/scale/np.pi.
            for i in range(nseg):
                indx, indy, distedge = fillPolygon(hx[:, i], hy[:, i],
                                                   i0 - ix0[i], j0 - iy0[i],
                                                   scale, gap * 0., segdiam,
                                                   index=1)
                pupil[indx + ix0[i], indy + iy0[i]] = attribute[i] * (
                            1. - (gap / scale / np.pi) / (
                                1 + (distedge / scale) ** 2))
        else:
            # Hard gaps
            for i in range(nseg):
                indx, indy, distedge = fillPolygon(hx[:, i], hy[:, i],
                                                   i0 - ix0[i], j0 - iy0[i],
                                                   scale, gap, segdiam, index=1)
                pupil[indx + ix0[i], indy + iy0[i]] = attribute[i]
    else:
        # attribute is [piston, tip, tilt]
        minimap = np.zeros((segdiam, segdiam))
        xmap = np.arange(segdiam) - segdiam / 2
        xmap, ymap = np.meshgrid(xmap, xmap, indexing='ij')  # [x,y] convention
        pitch = 1.244683637214  # diameter of inscribed circle
        diamseg = pitch * 2 / np.sqrt(3)  # diameter of circumscribed circle
        diamfrizou = (pitch + diamseg) / 2 * D / 40.  # average diameter
        # Calcul du facteur de mise a l'echelle pour l'unite des tilts.
        # xmap et ymap sont calculees avec un increment de +1 pour deux pixels
        # voisins, donc le facteur a appliquer est tel que l'angle se conserve
        # donc factunit*1 / scale = 4*factunit
        factunit = 4 * scale / diamfrizou
        for i in range(nseg):
            indx, indy, _ = fillPolygon(hx[:, i], hy[:, i], i0 - ix0[i],
                                        j0 - iy0[i], scale, 0., segdiam,
                                        index=1)
            minimap = attribute[0, i] + (factunit * attribute[1, i]) * xmap + (
                        factunit * attribute[2, i]) * ymap
            pupil[indx + ix0[i], indy + iy0[i]] = minimap[indx, indy]

    return pupil


#  _   _ ___ ____ _   _       _     _______     _______ _
# | | | |_ _/ ___| | | |     | |   | ____\ \   / / ____| |
# | |_| || | |  _| |_| |_____| |   |  _|  \ \ / /|  _| | |
# |  _  || | |_| |  _  |_____| |___| |___  \ V / | |___| |___
# |_| |_|___\____|_| |_|     |_____|_____|  \_/  |_____|_____|


def getEeltSegmentNumber():
    """
    Just returns the number of segments of the EELT nominal pupil, in order
    to be able to generate either reflectivities, or phase errors, or else.

    """
    hx, hy = generateCoordSegments(40., 0.)
    n = hx.shape[-1]
    return n


def generateEeltPupilMask(npt, dspider, i0, j0, pixscale, gap, rotdegree,
                          D=40.0, centerMark=0):
    """
    Generates a boolean pupil mask of the binary EELT pupil
    on a map of size (npt, npt).


    :returns: pupil image (npt, npt), boolean
    :param int npt: size of the output array
    :param float dspider: width of spiders in meters
    :param float i0, j0: index of pixels where the pupil should be centred.
                         Can be floating-point indexes.
    :param float pixscale: size of a pixel of the image, in meters.
    :param float gap: half-space between segments in meters
    :param float rotdegree: rotation angle of the pupil, in degrees.
    :param float D: diameter of the pupil. For the nominal EELT, D shall
                    be set to 40.0
    :param int centerMark: when centerMark!=0, a pixel is added at the centre of
        symmetry of the pupil in order to debug things using compass.
        centerMark==1 draws a point
        centerMark==2 draws 2 lines

    :Example:
    npt = 752
    i0 = npt/2+0.5
    j0 = npt/2+0.5
    rotdegree = 90.0
    pixscale = 40./npt
    dspider = 0.53
    gap = 0.02
    pup = generateEeltPupilMask(npt, dspider, i0, j0, pixscale, gap, rotdegree)

    """
    rot = rotdegree * np.pi / 180

    # Generation of segments coordinates.
    # hx and hy have a shape [6,798] describing the 6 vertex of the 798
    # hexagonal mirrors
    hx, hy = generateCoordSegments(D, rot)

    # From the data of hex mirrors, we build the pupil image using
    # boolean
    pup = generateSegmentProperties(True, hx, hy, i0, j0, pixscale, gap, npt, D)

    # SPIDERS ............................................
    nspider = 3  # for the day where we have more/less spiders ;-)
    if (dspider > 0 and nspider > 0):
        pup = pup & fillSpider(npt, nspider, dspider, i0, j0, pixscale, rot)

    # Rajout d'un pixel au centre (pour marquer le centre) ou d'une croix,
    # selon la valeur de centerMark
    if centerMark:
        pup = np.logical_xor(pup, centrePourVidal(npt, i0, j0, centerMark))

    return pup


def generateEeltPupilReflectivity(refl, npt, dspider, i0, j0, pixscale, gap,
                                  rotdegree, D=40.0, softGap=False):
    """
    Generates a map of the reflectivity of the EELT pupil, on an array
    of size (npt, npt).

    :returns: pupil image (npt, npt), with the same type of input argument refl
    :param float/int/bool refl: scalar value or 1D-array of the reflectivity of
           the segments.
           If refl is scalar, the value will be replicated for all segments.
           If refl is a 1D array, then it shall contain the reflectivities
           of all segments.
           On output, the data type of the pupil map will be the same as refl.
    :param int npt: size of the output array
    :param float dspider: width of spiders in meters
    :param float i0, j0: index of pixels where the pupil should be centred.
                         Can be floating-point indexes.
    :param float pixscale: size of a pixel of the image, in meters.
    :param float gap: half-space between segments in meters
    :param float rotdegree: rotation angle of the pupil, in degrees.
    :param float D: diameter of the pupil. For the nominal EELT, D shall
                    be set to 40.0
    :param bool softGap: if False, the gap between segments is binary 0/1
          depending if the pixel is within the gap or not. If True, the gap
          is a smooth region of a fwhm of 2 pixels with a depth related to the
          gap width.


    :Example:

    refl = np.ones(798)+np.random.randn(798)/20.
    dead = 3
    refl[(np.random.rand(dead)*797).astype(int)] = 0.
    npt = 1200
    i0 = npt/2+0.5
    j0 = npt/2+0.5
    rotdegree = 14.0
    pixscale = 44./npt
    dspider = 0.53
    gap = 0.02
    pup = generateEeltPupilReflectivity(refl, npt, dspider, i0, j0, pixscale,
                                        gap, rotdegree, softGap=True)

    """
    rot = rotdegree * np.pi / 180

    # Generation of segments coordinates.
    # hx and hy have a shape [6,798] describing the 6 vertex of the 798
    # hexagonal mirrors
    hx, hy = generateCoordSegments(D, rot)

    # From the data of hex mirrors, we build the pupil image according
    # to the properties defined by input argument <refl>
    pup = generateSegmentProperties(refl, hx, hy, i0, j0, pixscale, gap, npt, D,
                                    softGap=softGap)

    # SPIDERS ............................................
    nspider = 3  # for the day where we have more/less spiders ;-)
    if (dspider > 0 and nspider > 0):
        pup = pup * fillSpider(npt, nspider, dspider, i0, j0, pixscale, rot)

    return pup


def generateEeltPupilPhase(phase, npt, dspider, i0, j0, pixscale, rotdegree,
                           D=40.0):
    """
    Generates a map of the segments phase errors of the EELT pupil, on an array
    of size (npt, npt).

    :returns: phase image (npt, npt), with the same type of input argument phase
    :param float phase: scalar value or 2D-array of the piston, tip
           and tilt of the segments. The array shall be of dimension
           [3, 798] that contains [piston, tip, tilt]
    :param int npt: size of the output array
    :param float dspider: width of spiders in meters
    :param float i0, j0: index of pixels where the pupil should be centred.
                         Can be floating-point indexes.
    :param float pixscale: size of a pixel of the image, in meters.
    :param float rotdegree: rotation angle of the pupil, in degrees.
    :param float D: diameter of the pupil. For the nominal EELT, D shall
                    be set to 40.0

    :Example:

    phase = np.random.randn(3,798)
    phase = np.zeros((3,798)); phase[1,:]=1.
    npt = 752
    i0 = npt/2+0.5
    j0 = npt/2+0.5
    rotdegree = 90.0
    pixscale = 41./npt
    dspider = 0.51
    pup = generateEeltPupilPhase(phase, npt, dspider, i0, j0, pixscale,
                                 rotdegree)

    """
    rot = rotdegree * np.pi / 180

    # Generation of segments coordinates.
    # hx and hy have a shape [6,798] describing the 6 vertex of the 798
    # hexagonal mirrors
    hx, hy = generateCoordSegments(D, rot)

    # From the data of hex mirrors, we build the pupil phase image according
    # to the properties defined by input argument <phase>
    pup = generateSegmentProperties(phase, hx, hy, i0, j0, pixscale, 0.0, npt,
                                    D)

    return pup


"""
refl = np.ones(798)+np.random.randn(798)/10.
N = npt = 800
i0 = N/2+0.5
j0 = N/2+0.5
rotdegree = 10.0
scale = pixscale = 41./N
dspider = 0.51


#smap = generateEeltPupil_slow(N, dspider, i0, j0, scale, rotdegree)

p = generateEeltPupilMask(N, dspider, i0, j0+10, scale, rotdegree)
plt.clf()
plt.matshow(p, fignum=1)

#p = generateEeltPupilReflectivity(refl, N, dspider, i0, j0, pixscale, 
                                   rotdegree, D=40.0)

phase = np.zeros((3,798)); phase[1,:]=1.
phase = np.random.randn(3,798)
p = generateEeltPupilPhase(phase, N, dspider, i0, j0, pixscale, rotdegree, 
                           D=40.0)

plt.matshow(p, fignum=1)


N = 2048
i0 = N/2
j0 = N/2
rotdegree = 10.0
scale = pixscale = 40./800
dspider = 0.53

pup = generateEeltPupilMask(N, dspider, i0, j0, scale, rotdegree)
phase = np.random.randn(3,798) * 1.  # en microns rms
phase = np.zeros((3,798)); phase[1,:]=0.3
delta = generateEeltPupilPhase(phase, N, dspider, i0, j0, pixscale, rotdegree)
lam = 1.65  # microns
F = np.exp((1j*2*np.pi/lam)*delta) * pup
psf = np.fft.fftshift(np.abs(np.fft.fft2(F))) / np.sum(pup)


"""



import pytest

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

from astropy.io import fits

from anisocado.psf import AnalyticalScaoPsf

PLOTS = False


class TestInit:
    def test_initialises_with_nothing(self):
        assert isinstance(AnalyticalScaoPsf(), AnalyticalScaoPsf)

    def test_initialises_with_bunch_of_keywords(self):
        psf = AnalyticalScaoPsf(N=256, wavelengthIR=1.2e-6)
        assert isinstance(psf, AnalyticalScaoPsf)

    def test_throws_warning_if_keyword_not_spelt_correctly(self):
        psf = AnalyticalScaoPsf(N=256, wavelengthIR=1.2e-6, hello="world")
        assert isinstance(psf, AnalyticalScaoPsf)

    def test_on_axis_psf_is_made(self):
        psf = AnalyticalScaoPsf(N=512)
        assert isinstance(psf.psf_on_axis, np.ndarray)

        if PLOTS:
            plt.imshow(psf.psf_on_axis.T, origin="lower", norm=LogNorm(), vmin=1e-7)
            plt.show()

    def test_kernel_sums_to_one(self):
        psf = AnalyticalScaoPsf(N=256)
        assert np.sum(psf.psf_latest) == pytest.approx(1)


class TestShiftPSF:
    def test_psf_blurs_when_shifted(self):
        psf = AnalyticalScaoPsf(N=1024)
        sr_orig = psf.strehl_ratio

        psf.shift_off_axis(0, 10)
        sr_last = psf.strehl_ratio

        assert sr_last < sr_orig

        if PLOTS:
            plt.subplot(121)
            plt.imshow(psf.psf_on_axis.T, origin="lower", norm=LogNorm(), vmin=3e-6)
            plt.colorbar()

            plt.subplot(122)
            plt.imshow(psf.psf_latest.T, origin="lower", norm=LogNorm(), vmin=3e-6)
            plt.colorbar()
            plt.show()

    def test_kernel_sums_to_one(self):
        psf = AnalyticalScaoPsf(N=512)
        psf.shift_off_axis(0, 0)
        assert np.sum(psf.psf_latest) == pytest.approx(1)


class TestHDUProperty:
    def test_returns_fits_imagehdu(self):
        psf = AnalyticalScaoPsf(N=512)
        psf.shift_off_axis(10, 10)
        hdu = psf.hdu
        assert isinstance(hdu, fits.ImageHDU)
        assert hdu.header["CDELT1"] == 4 / (3600 * 1000.)
        assert hdu.header["CRVAL1"] == 10 / 3600.

        if PLOTS:
            plt.imshow(hdu.data.T, origin="l", norm=LogNorm())
            plt.show()






"""Originally from file _anisocado.py"""

import numpy
import numpy as np
from . import pupil_utils

#  ____  _____    _    ____  __  __ _____
# |  _ \| ____|  / \  |  _ \|  \/  | ____|
# | |_) |  _|   / _ \ | | | | |\/| |  _|
# |  _ <| |___ / ___ \| |_| | |  | | |___
# |_| \_\_____/_/   \_\____/|_|  |_|_____|

"""
Hello.
This files contains some useful functions related to psf generation.
They are just raw, so that you can insert them into your own classes
as desired.
The functions are commented to help you understand what's going on, when
you need to modify them.

Don't read that code anyway.
Go right now to [the file _anisocado.py].
It contains examples that will show you how to use the functions, what they do,
etc.
"""


def defineDmFrequencyArea(kx, ky, rotdegree, dactu=0.5403):
    """
    <kx>        : spatial frequency produced by the
                  function computeSpatialFreqArrays() (metres^-1)
    <ky>        : idem kx
    <rotdegree> : rotation of M4 (degrees)
    <dactu>     : value of actuator pitch of M4 (metres)

    The function returns the 2D domain of spatial frequencies which can be
    impacted by M4, i.e. the M4 compensation domain.
    Underlying assumtions are than M4 is of an hexagonal pattern, with an
    inter-actuator distance <dactu> set by default to 54.03 cm.

    Result is returned with a frequency corner-centred representation.

    Standalone example::

        N = 512               # output will be 512x512
        pixelSize = 4.2       # 4.2 mas
        wavelength = 1.65e-6  # metres
        kx, ky = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, 0)
        plt.imshow(np.fft.fftshift(M4).T, origin='l')

    """

    # cut-off frequency
    fc = (1./np.sqrt(3)) / dactu

    # mask frequency definition
    A = np.pi/3  # 60 degrees
    A0 = rotdegree * np.pi / 180
    msk = np.abs(np.cos(A0)*ky+np.sin(A0)*kx)<fc
    msk = np.logical_and(msk, np.abs(np.cos(A+A0)*ky+np.sin(A+A0)*kx)<fc)
    msk = np.logical_and(msk, np.abs(np.cos(2*A+A0)*ky+np.sin(2*A+A0)*kx)<fc)
    k = np.sqrt(kx**2 + ky**2)
    msk = np.logical_and(msk, k < (fc*1.115))

    return msk


def computeSpatialFreqArrays(N, pixelSize, wavelength):
    """
    <N>             : size of the output image (NxN)
    <pixelSize>     : size of the pixels of the psf image (mas)
    <wavelength>    : wavelength (metres)

    The function returns a tuple of spatial frequencies (m^-1) in X and Y
    together with the pixel scale of these (the value will be useful later
    on in other procedures).
    Results of 2D arrays are returned with a Fourier corner-centred
    representation.

    Standalone example::

        N = 512               # output will be 512x512
        pixelSize = 4.2       # 4.2 mas
        wavelength = 1.65e-6  # metres
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)

    """

    # array of indices centred 'as expected' by Fourier frequencies, in 1D
    k1d = np.fft.fftshift(np.arange(N) - (N//2))

    # Proper scaling to transform indices in spatial frequencies.

    # dX = pixelSize * 4.84813681109536e-09  # from mas to radians
    dX = pixelSize * 4.84813681109536e-06  # from arcsec to radians
    uk = dX / wavelength
    k1d = k1d * uk  # now this is a spatial freq in metres^-1

    # now creating 2D arrays of spatial frequency
    kx, ky = np.meshgrid(k1d, k1d, indexing='ij') # for convention [x,y]

    return kx, ky, uk


def computeWiener(kx, ky, L0, r0):
    """
    <kx>  : spatial frequency produced by
            the function computeSpatialFreqArrays() (metres^-1)
    <ky>  : idem kx
    <L0>  : value of the outer scale (metres)
    <r0>  : value of Fried parameter r0 (metres)

    The function returns the 2D spectrum of the turbulence with an outer
    scale L0. It is expressed in rad^2.m^2.

    Result is returned with a frequency corner-centred representation.

    Standalone example::

        N = 512               # output will be 512x512
        pixelSize = 4.2       # 4.2 mas
        wavelength = 1.65e-6  # metres.  H band.
        L0 = 25.              # 25 metres outer scale
        r0 = 0.6              # r0 = 60cm in H band
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        W = computeWiener(kx, ky, L0, r0)

    """

    # computation of Wiener spectrum expressed in radians^2 (at the wavelength
    # where r0(lambda) is expressed !)
    Wiener = (kx**2 + ky**2 + 1./L0**2.)**(-11./6)
    Wiener *= 0.0228956 * r0**(-5./3)

    # frequency 0 set to 0. It's wrong anyway.
    Wiener[0, 0] = 0.

    return Wiener


def anisoplanaticSpectrum(Cn2h, layerAltitude, L0, offx, offy, wavelength,
                          kx, ky, Wiener, M4):
    """
    <Cn2h>          : list of normalised (np.sum(Cn2h)==1.0) strengh of
                      turbulence of each layer (no unit)
    <layerAltitude> : list of altitudes of the turbulence layers (metres)
    <L0>            : value of the outer scale (metres)
    <offx>          : off-axis distance of the star along X axis (arcsec)
    <offy>          : idem along Y axis (arcsec)
    <wavelength>    : wavelength (metres)
    <kx>            : spatial frequency produced by
                      the function computeSpatialFreqArrays() (metres^-1)
    <ky>            : idem kx
    <Wiener>        : turbulent spectrum from function computeWiener()
    <M4>            : frequency domain of M4 (boolean) computed by the
                      function M4 = defineDmFrequencyArea(kx, ky, 0)

    Computes the phase spectrum (spatial frequencies) contribution due to the
    anisoplanatism in the AO compensation for a source located off-axis from
    the SCAO guide star by some amount (offx, offy).
    In input, the distribution of the turbulence in altitude is given.

    Result is returned with a frequency corner-centred representation.

    Example::

        N = 512
        pixelSize = 4.2
        Cn2h = [0.3, 0.2, 0.2, 0.1]
        layerAltitude = [0,1000,8000,15000]
        L0 = 25.0
        offx, offy = (34., 12.)
        wavelength = 1.65e-6
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, 0)
        W = computeWiener(kx, ky, L0, 1.0)
        f = anisoplanaticSpectrum(Cn2h, layerAltitude, L0, offx, offy,
                                  wavelength, kx, ky, W, M4)

    """

    # number of turbulent layers involved in that computation
    nlayers = len(Cn2h)

    # conversion arcsec to radians
    RASC = 206264.8062471

    # prepare memory alloc for transfer function of anisoplanatism
    Haniso = np.zeros(kx.shape)

    # loop over turbulent layers, summing transfer function of each layer
    for i in range(nlayers):
        dx = layerAltitude[i] * offx / RASC # shift in metres on the layer in X
        dy = layerAltitude[i] * offy / RASC # idem, in Y
        tmp = (2j*np.pi*dx)*kx + (2j*np.pi*dy)*ky
        Haniso += Cn2h[i] * np.abs(1 - np.exp( tmp ))**2

    # now applying the transfer function on the Wiener spectrum, only in the
    # spatial frequency range of M4
    Waniso = np.zeros(Haniso.shape)
    Waniso[M4] = Haniso[M4] * Wiener[M4]

    return Waniso


def fittingSpectrum(Wiener, M4):
    """
    <Wiener>        : turbulent spectrum from function computeWiener()
    <M4>            : frequency domain of M4 (boolean) computed by the
                      function M4 = defineDmFrequencyArea(kx, ky, 0)

    Returns the spatial spectrum of the (so-called) fitting error, i.e. the
    residual phase after a full, perfect, ideal, instantaneous, super-duper,
    hyper-clean, theoretical compensation of M4. It is expressed in rad^2.m^2.
    Result is returned with a frequency corner-centred representation.

    Example::

        N = 512
        pixelSize = 4.2
        L0 = 25.0
        offx, offy = (34., 12.)
        wavelength = 1.65e-6
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, 0)
        r0 = 0.6
        W = computeWiener(kx, ky, L0, r0)
        f = fittingSpectrum(W, M4)

    """

    Wfit = Wiener.copy()
    Wfit[M4] = 0.0 # M4 cancels whatever is in its compensation domain
    return Wfit


def otherSpectrum(nmRms, M4, uk, wavelength):
    """
    <nmRms>  : number of nm rms
    <M4>     : frequency domain of M4 (boolean) computed by the
                      function M4 = defineDmFrequencyArea(kx, ky, 0)
    <uk> : # size of the 'spatial frequency pixel' in m^-1
    <wavelength>    : wavelength (metres)

    Example::

        N = 512
        pixelSize = 4.2
        L0 = 25.0
        offx, offy = (34., 12.)
        wavelength = 1.65e-6
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, 0)
        nmRms = 250.
        f = otherSpectrum(nmRms, M4, uk, wavelength)

    """

    fact = 2 * np.pi * nmRms * 1e-9 / uk / wavelength
    fact = fact**2
    tot = np.sum(M4)
    Wothers = np.zeros(M4.shape)
    Wothers[M4] = fact / tot

    return Wothers


def aliasingSpectrum(kx, ky, r0, L0, M4, dssp=0.4015):
    """
    Example::

        N = 512
        pixelSize = 4.2
        L0 = 25.0
        r0 = 0.6
        wavelength = 1.65e-6
        rotdegree = 10.
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, rotdegree)
        W = aliasingSpectrum(kx, ky, r0, L0, M4)

    """

    ke = 1.0 / dssp  # computes the sampling spatial-frequency of the WFS

    kxt = kx[M4]
    kyt = ky[M4]
    Wt  = ((kxt-ke)**2 + kyt**2 + 1./L0**2.)**(-11./6)
    Wt += ((kxt+ke)**2 + kyt**2 + 1./L0**2.)**(-11./6)
    Wt += (kxt**2 + (kyt-ke)**2 + 1./L0**2.)**(-11./6)
    Wt += (kxt**2 + (kyt+ke)**2 + 1./L0**2.)**(-11./6)
    Wt *= 0.0228956 * r0**(-5./3)
    Walias = np.zeros(M4.shape)
    Walias[M4] = Wt

    return Walias


def computeBpSpectrum(kx, ky, V, Fe, tret, gain, Wiener, M4):
    """
    Example::

        N = 512
        pixelSize = 4.2
        L0 = 25.0
        r0 = 0.6
        wavelength = 1.65e-6
        rotdegree = 10.
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, rotdegree)
        W = computeWiener(kx, ky, L0, r0)

        V = 10.
        Fe = 500.
        tret = 0.004
        gain = 0.3
        f = computeBpSpectrum(kx, ky, V, Fe, tret, gain, W, M4)

    """

    k = np.sqrt(kx*kx + ky*ky)
    nu = k * V / np.sqrt(2)    # pourquoi un sqrt(2) ? je ne saurais dire ...!!!
    Wbp = hcor(nu, Fe, tret, gain, 500) * Wiener
    Wbp[np.logical_not(M4)] = 0.
    return Wbp


def hcor(freq, Fe, tret, G, BP, an=True):
    """
    ***** Fonction extraite de STYC le 9 Jan 2013 *****

    The option an=1 sets the integrator to "analog". Doing this, an
    extra 1/2 frame delay is added compared to case of the numeric
    integrator an=0.

    <tret> is the delay expressed as a *time in seconds*, between the
    end of the integration and the start of the command.

    """

    Te = 1. / Fe
    p = 1j * 2 * np.pi * freq + 1e-12

    Hint = 1./(1-np.exp(-p*Te)) # numeric integrator
    Hccd = (1.-np.exp(-p*Te))/(p*Te)  # echant bloqueur avec retard 1/2 trame
    Hdac = Hccd                    # echant bloqueur avec retard 1/2 trame
    Hret = np.exp(-p*tret)
    Hmir = 1./(1. + 1j*freq/BP)
    Hbo = Hint * Hccd * Hdac * Hret * Hmir
    Hcor   = 1./abs(1 + Hbo*G)**2

    return Hcor


def convertSpectrum2Dphi(W, uk):
    """
    <W>  : spatial spectrum to be converted into phase structure function
           in rd^2.m^2
    <uk> : # size of the 'spatial frequency pixel' in m^-1

    Converts the 2D spectrum into a phase structure function Dphi.
    Uses Dphi(r) = $ $ (1-cos(2.pi.k.r)) W(k) d2k
    Computation of Dphi is in radians^2 at the wavelength of r0.
    """

    W[0, 0] = 0.0
    W[0, 0] = -np.sum(W)
    Dphi = 2*np.abs(np.fft.fft2(W)) *  (uk**2)
    return Dphi


def fake_generatePupil(N, deadSegments, rotdegree, pixelSize, wavelength):
    """
    <N>            : size of the output image, that is made to match the size
                     of the (square) psf image to be processed. In other
                     words, N = psf.shape[0]
    <deadSegments> : number of hexa segments of M1 that are missing
    <rotdegree>    : pupil rotation in degrees
    <pixelSize>    : size of the pixels of the psf image (mas)
    <wavelength>   : wavelength (metres)

    Examples::

        N = 512
        deadSegments = 3
        rotdegree = 14.
        pixelSize = 4.2
        wavelength = 1.65e-6
        pup = fake_generatePupil(N, deadSegments, rotdegree, pixelSize,
                                 wavelength)

    """
    nseg = pupil_utils.getEeltSegmentNumber()
    refl = np.ones(nseg)+np.random.randn(nseg)/20.
    if deadSegments:
        refl[(np.random.rand(deadSegments)*nseg).astype(int)] = 0.
    i0 = N/2+0.5
    j0 = N/2+0.5

    # field of view of the psf image in rd
    FoV = N * pixelSize * 4.84813681109536e-06  # from arcsec to radians
    # original line used mas
    # FoV = N * pixelSize * 4.84813681109536e-09  # from mas to radians

    # pixel scale of pupil image
    pixscale = wavelength / FoV   # expressed in metres
    dspider = 0.53
    gap = 0.02
    pup = pupil_utils.generateEeltPupilReflectivity(refl, N, dspider, i0, j0,
                                                    pixscale, gap, rotdegree,
                                                    softGap=True)
    return pup


def computeEeltOTF(pup):
    """
    """
    # Computation of telescope OTF
    Nx, Ny = pup.shape
    FTOtel = np.fft.fft2( np.abs(np.fft.fft2(pup))**2 ).real
    FTOtel /= np.sum(pup)**2 * Nx * Ny
    return FTOtel


def core_generatePsf(Dphi, FTOtel):
    """
    Examples
    --------
    ::
        N = 1024
        pixelSize = 4.2

        # atmospheric profile (old ESO profile before 2010)
        layerAltitude = [47., 140, 281, 562, 1125, 2250, 4500, 9000, 18000.]

        # from ref. E-SPE-ESO-276-0206_atmosphericparameters
        Cn2h = [52.24, 2.6, 4.44, 11.60, 9.89, 2.95, 5.98, 4.30, 6]
        Cn2h = np.array(Cn2h)
        Cn2h /= np.sum(Cn2h)

        # outer scale
        L0 = 25.0
        offx, offy = (15., 20.)
        wavelength = 1.65e-6
        rotdegree = 10.0
        kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelength)
        M4 = defineDmFrequencyArea(kx, ky, rotdegree)
        r0 = 0.6

        # This is the turbulent spectrum ....
        W = computeWiener(kx, ky, L0, r0)

        # And here are some of the PSF-destroyers
        Waniso = anisoplanaticSpectrum(Cn2h, layerAltitude, L0, offx, offy,
                                       wavelength, kx, ky, W, M4)
        Wfit = fittingSpectrum(W, M4)
        nmRms = 100.
        Wother = otherSpectrum(nmRms, M4, uk, wavelength)
        Dphi = convertSpectrum2Dphi(Waniso + Wfit + Wother, uk)

        # Here, i need to generate a kind of PSF or telescope OTF
        deadSegments = 3
        pup = fake_generatePupil(N, deadSegments, rotdegree, pixelSize,
                                 wavelength)
        FTOtel = computeEeltOTF(pup)

        psf = core_generatePsf(Dphi, FTOtel)

        print(psf.max())
        window = 100
        plt.imshow( psf[N//2-window:N//2+window, N//2-window:N//2+window]**0.3 )

    """

    # total FTO
    FTO = np.exp(-0.5*Dphi) * FTOtel

    # PSF
    psf = np.fft.fftshift( np.fft.fft2(FTO).real )
    return psf


def createAdHocScaoPsf(N, pixelSize, wavelengthIR, rotdegree, r0Vis, nmRms):
    """
    <N>            : size of the output image, that is made to match the size
                     of the (square) psf image to be processed. In other
                     words, N = psf.shape[0]
    <pixelSize>    : size of the pixels of the psf image (mas)
    <wavelengthIR> : IR wavelength (metres)
    <rotdegree>    : pupil rotation (degrees)
    <r0Vis>        : value of r0 in the visible (metres)
    <nmRms>        : number of nm rms affecting the psf (nm)

    Do not use that function.
    It's just there to create a quicky random shitty psf.
    It's there because i needed a random shitty psf.
    So i wrote it.
    And it's still there.

    Example::

        N = 512
        pixelSise = 4.2         # mas
        wavelengthIR = 1.65e-6  # metres
        rotdegree = 10.
        r0Vis = 0.12
        nmRms = 150.
        psf, pup = createAdHocScaoPsf(N, pixelSize, wavelengthIR, rotdegree,
                                      r0Vis, nmRms)

    """

    # let's compute r0 in the IR using the
    # r0 chromatic translation formula
    wavelengthVis = 500e-9
    r0IR = r0Vis * (wavelengthIR / wavelengthVis)**(6/5.)

    kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelengthIR)
    M4 = defineDmFrequencyArea(kx, ky, rotdegree)

    # I hardcode an outer scale of 25 metres
    L0 = 25.

    # This is the turbulent spectrum ....
    W = computeWiener(kx, ky, L0, r0IR)

    # And here are some of the PSF-destroyers
    Wfit = fittingSpectrum(W, M4)
    nmRms = 200.
    Wother = otherSpectrum(nmRms, M4, uk, wavelengthIR)
    Dphi = convertSpectrum2Dphi(Wfit + Wother, uk)

    # Here, i need to generate a kind of PSF or telescope OTF
    deadSegments = 3
    pup = fake_generatePupil(N, deadSegments, rotdegree, pixelSize,
                             wavelengthIR)
    FTOtel = computeEeltOTF(pup)

    psf = core_generatePsf(Dphi, FTOtel)

    return psf, pup


def r0Converter(r0, lambda1, lambda2):
    """
    Converts a r0 defined at some wavelength lambda1,
    into a r0' at lambda2.
    Lambda1 and 2 shall be in the SAME unit.
    Returned r0 will be in the SAME unit than the input one.

    Example::

        r0_K = r0Converter(0.12, 500, 2200)

    """
    return r0 * (lambda2/lambda1)**(6/5.)


def airmassImpact(r0_at_zenith, zenith_distance):
    """
    <r0_at_zenith>    : r0 at zenith (any unit is ok)
    <zenith_distance> : zenith distance (degrees)

    The seeing/r0 actually perceived by a telescope depends on the zenith
    distance. This function converts a r0 given at zenith into the real r0
    actually observed by the telescope.
    """

    z = zenith_distance * np.pi / 180  # the same, in radians
    r0 = r0_at_zenith * np.cos(z)**(3./5)

    return r0


def get_atmospheric_turbulence(myProfile='EsoMedian'):
    """
    Returns the relative level of turbulence at a given height

    Note: The np.sum(Cn2h) = 1.0
    Turbulence profile have been taken from ESO-258292: "Relevant Atmospheric
    Parameters for E-ELT AO Analysis and Simulations".

    The following 3 turbulence profile are currently available:

    * ``EsoQ1``

        Best atmospheric conditions - First Quartile Armazones atmospheric
        turbulence profile

    * ``EsoMedian`` [also ``officialEsoMedian``]

        Median atmospheric conditions - Median Armazones atmospheric turbulence
        profile

    * ``EsoQ4``

        Worst atmospheric conditions - First Quartile Armazones atmospheric
        turbulence profile

    Additional turbulence profiles include:

    * ``oldEso``

        The old ESO profile from before 2010. Cn2h paramateres are taken from
        ref. E-SPE-ESO-276-0206_atmosphericparameters

    * ``gendron``

        The Gendron profile. Short, fast. Saves CPU. Carbon efficient.
        The np.sum(Cn2h) = 1.0 is hyper-guaranteed here.


    Parameters
    ----------
    myProfile : str, optional
        Profile name: ['EsoQ1', 'EsoMedian', 'EsoQ4', 'oldEso', 'gendron']

    Returns
    -------
    layerAltitude : list of floats
        [m] height of layer above ground
    Cn2h : list of floats
        Relative strength of turbulence

    """

    layerAltitude, Cn2h = [], []

    if myProfile == 'oldEso':
        layerAltitude = [47., 140, 281, 562, 1125, 2250, 4500, 9000, 18000.]
        Cn2h = [0.5224, 0.026, 0.0444, 0.116, 0.0989,
                0.0295, 0.0598, 0.043, 0.06]

    elif myProfile == 'officialEsoMedian' or myProfile == 'EsoMedian' :
        layerAltitude = [30, 90, 150, 200, 245, 300, 390, 600, 1130, 1880, 2630,
                         3500, 4500, 5500, 6500, 7500, 8500, 9500, 10500, 11500,
                         12500, 13500, 14500, 15500, 16500, 17500, 18500, 19500,
                         20500, 21500, 22500, 23500, 24500, 25500, 26500]
        Cn2h = [24.2, 12, 9.68, 5.9, 4.73, 4.73, 4.73, 4.73, 3.99, 3.24, 1.62,
                2.6, 1.56, 1.04, 1, 1.2, 0.4, 1.4, 1.3, 0.7, 1.6, 2.59, 1.9,
                0.99, 0.62, 0.4, 0.25, 0.22, 0.19, 0.14, 0.11, 0.06, 0.09, 0.05,
                0.04]
        Cn2h = np.array(Cn2h)
        Cn2h /= np.sum(Cn2h)

    elif myProfile == 'EsoQ1':
        layerAltitude = [30, 90, 150, 200, 245, 300, 390, 600, 1130, 1880, 2630,
                         3500, 4500, 5500, 6500, 7500, 8500, 9500, 10500, 11500,
                         12500, 13500, 14500, 15500, 16500, 17500, 18500, 19500,
                         20500, 21500, 22500, 23500, 24500, 25500, 26500]
        Cn2h = [22.6, 11.2, 10.1, 6.4, 4.15, 4.15, 4.15, 4.15, 3.1, 2.26, 1.13,
                2.21, 1.33, 0.88, 1.47, 1.77, 0.59, 2.06, 1.92, 1.03, 2.3, 3.75,
                2.76, 1.43, 0.89, 0.58, 0.36, 0.31, 0.27, 0.2, 0.16, 0.09, 0.12,
                0.07, 0.06]
        Cn2h = np.array(Cn2h)
        Cn2h /= np.sum(Cn2h)

    elif myProfile == 'EsoQ4':
        layerAltitude = [30, 90, 150, 200, 245, 300, 390, 600, 1130, 1880, 2630,
                         3500, 4500, 5500, 6500, 7500, 8500, 9500, 10500, 11500,
                         12500, 13500, 14500, 15500, 16500, 17500, 18500, 19500,
                         20500, 21500, 22500, 23500, 24500, 25500, 26500]
        Cn2h = [23.6, 13.1, 9.81, 5.77, 6.58, 6.58, 6.58, 6.58, 5.4, 3.2, 1.6,
                2.18, 1.31, 0.87, 0.37, 0.45, 0.15, 0.52, 0.49, 0.26, 0.8, 1.29,
                0.95, 0.49, 0.31, 0.2, 0.12, 0.1, 0.09, 0.07, 0.06, 0.03, 0.05,
                0.02, 0.02]
        Cn2h = np.array(Cn2h)
        Cn2h /= np.sum(Cn2h)

    elif myProfile == 'gendron':
        Cn2h = [1.0]
        layerAltitude = [4414.]

    return layerAltitude, Cn2h


def get_profile_defaults(myProfile="EsoMedian"):
    """
    Data taken from ESO-258292

    Parameters
    ----------
    myProfile : str
        [EsoMedian, EsoQ1, EsoQ4, OldEso, Gendron]

    Returns
    -------
    seeing : float
        [arcsec]
    zen_dist : float
        [deg] Zenith distance
    wind_alpha : float
        Multiplication factor for the wind profile

    """
    seeing = 0.67
    zen_dist = 30
    wind_alpha = 1.
    if myProfile == 'EsoQ1':
        seeing = 0.4
        zen_dist = 0
        wind_alpha = 0.88
    elif myProfile == 'EsoQ4':
        seeing = 1.0
        zen_dist = 60
        wind_alpha = 1.3

    return seeing, zen_dist, wind_alpha


def clean_psf(psf, threshold):
    psf[psf < threshold] = 0
    edge_threshold = np.median([psf[:, 0], psf[0, :]])
    psf[psf < edge_threshold] = 0
    psf /= np.sum(psf)

    return psf


def round_edges(kernel, edge_width=10):
    n = edge_width
    falloff = np.cos(1.5708 * np.arange(n) / (n-1)).reshape([1, n])

    kernel[:n, :] *= falloff.T[::-1, :]
    kernel[-n:, :] *= falloff.T
    kernel[:, :n] *= falloff[:, ::-1]
    kernel[:, -n:] *= falloff

    return kernel






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 12 18:48:47 2019

@author: Eric Gendron
Edited by Kieran Leschinski
"""

import matplotlib.pyplot as plt
from anisocado.psf_utils import *
from anisocado.psf_utils import get_atmospheric_turbulence

#  _   _             ____                  _
# | | | |___  ___   / ___|__ _ ___  ___   / |
# | | | / __|/ _ \ | |   / _` / __|/ _ \  | |
# | |_| \__ \  __/ | |__| (_| \__ \  __/  | |
#  \___/|___/\___|  \____\__,_|___/\___|  |_|


def shift_scao_psf(plots=False):
    """
    Problem:
    You have an on-axis PSF.
    You want to 'move' it off-axis, let's say (+15, +20) arcsec.


    For that, you will need to know:
        - the Cn2h profile
        - the global r0
        - the global L0

    """
    ###########
    # Setup

    # Let's take an example. I create a PSF, that could be coming from
    # a SCAO simulation. It has 512x512 pixels, sampled with 4.2 mas in H band.
    # Pupil is rotated by 10 deg. The r0 was 12cm.
    #
    # That psf will be the starting point.
    N = 512
    pixelSize = 4.2         # mas
    wavelengthIR = 1.65e-6  # metres
    rotdegree = 10.         # deg
    r0Vis = 0.12
    nmRms = 150.
    psf, pup = createAdHocScaoPsf(N, pixelSize, wavelengthIR, rotdegree, r0Vis,
                                  nmRms)

    if plots:
        # I can even look at it:
        plt.imshow(psf.T, origin='l')
        print('Strehl ratio of initial psf is ', psf.max())

    # OK. That's the starting point.......................

    # Now I need to know the atmospheric properties, in particular the Cn2h
    # profile. Let me offer you a little selection of atmospheric profiles.
    profile_name = "gendron"
    layerAltitude, Cn2h = get_atmospheric_turbulence(profile_name)

    # Let us define the outer scale value. 25 metres is the Armazones median
    # value from doc. ESO-258292.
    # Choose 15 m for a lucky observer. 50 m for the looser.
    L0 = 25.   # 25 metres is the Armazones median value from doc. ESO-258292.

    # Now, the seeing.
    # Here, we're cheating, we already know the r0 is 12cm because we've
    # generated the PSF with this.
    r0Vis = 0.12  # I know, we know it already ...

    # I also need to know where's the off-axis star I want to simulate
    # in arcsecs (this one is for those who'd like to check nothing will change
    # at the end)
    # offx, offy = (0, 0)
    offx, offy = (0., 16.)   # in arcsecs

    ####################
    # Generate PSF

    # Then let's start the work. I will create spatial frequency arrays.
    kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelengthIR)

    # convert r0 in the infra-red
    r0IR = r0Converter(r0Vis, 500e-9, wavelengthIR)

    # and create M4 working domain
    M4 = defineDmFrequencyArea(kx, ky, rotdegree)

    # and finally the turbulent spectrum ....
    W = computeWiener(kx, ky, L0, r0IR)

    # and all this will be used to run that function below, that will compute
    #  the spatial spectrum of the phase due to anisoplanatism
    Waniso = anisoplanaticSpectrum(Cn2h, layerAltitude, L0, offx, offy,
                                   wavelengthIR, kx, ky, W, M4)

    # Transforming this spectrum into a phase structure function
    Dphi = convertSpectrum2Dphi(Waniso, uk)

    # Here, the on-axis psf comes into play ... I take its Fourier transform
    fto = np.fft.fft2(np.fft.fftshift(psf)) / N**2  # it's complex.

    psf_aniso = core_generatePsf(Dphi, fto)

    print('Strehl off-axis is', psf_aniso.max())
    plt.imshow(psf_aniso.T, origin='l' )

    return psf_aniso

#  _   _             ____                 ____
# | | | |___  ___   / ___|__ _ ___  ___  |___ \
# | | | / __|/ _ \ | |   / _` / __|/ _ \   __) |
# | |_| \__ \  __/ | |__| (_| \__ \  __/  / __/
#  \___/|___/\___|  \____\__,_|___/\___| |_____|


def exnihilo_scao_psf():
    """

    You want to generate SCAO PSFs ex-nihilo, using a simple, approximative
    simulation software.
    You are aware that the simulated PSFs will be perfectly smooth (infinitely
    converged), they do not reflect the fluctuations associated to short
    exposures.

    For this you need to know all the parameters of the simulation.

    """

    ###########
    # Setup

    # nb of pixels of the image to be simulated.
    N = 1024
    pixelSize = 3.1  # mas

    # this is where you observe
    wavelengthIR = 2.2e-6   # oh! a K band ...

    # Now you need to tell what turbulence looks like.
    # Here you can imagine to take those numbers into ESO doc, statistics,
    # you can also include some dependence of r0 with respect to airmass
    #
    # If everything is as usual, you should end up with too many parameters that
    # nobody knows about, and someone will tell you "ok it's very nice, but can
    # you please simplify this ?"
    layerAltitude = [47., 140, 281, 562, 1125, 2250, 4500, 9000, 18000.]
    # from ref. E-SPE-ESO-276-0206_atmosphericparameters
    Cn2h = [0.5224, 0.026, 0.0444, 0.116, 0.0989, 0.0295, 0.0598, 0.043, 0.06]
    L0 = 25.   # 25 metres is the Armazones median value
    seeing = 0.8          # in arcseconds
    r0Vis = 0.103 / seeing  # r0Vis is in metres here, 0.103 is in metres.arcsec
    r0IR = r0Converter(r0Vis, 500e-9, wavelengthIR) # convert r0 at 500nm to IR

    # Just to use that wonderful function, i decide that the seeing given here
    # was expressed at zenith, while our telescope observes at 30° from zenith.
    # This will transform our r0 into the effective/actual one seen at 30°.
    # In addition, the turbulent layers will all appear further away from the
    # telescope, so that their apparent distance grows with airmass --> this is
    # very bad for anisoplanatism !..
    zenDist = 30.  # I observe at 30° from zenith
    r0IR = airmassImpact(r0IR, zenDist)  # apparent seeing degrades with airmass

    layerAltitude = np.array(layerAltitude)
    layerAltitude *= 1/np.cos(zenDist*np.pi/180)  # layers appear further away

    # Also you may want to say something about how the pupil is rotated wrt
    # your image
    rotdegree = 10.0

    # And you also need to generate the EELT pupil properly
    deadSegments = 5  # there are some missing segments tonight !
    pup = fake_generatePupil(N, deadSegments, rotdegree, pixelSize,
                             wavelengthIR)

    # For temporal aspects you need to know the characteristics of your system
    V = 10.        # wind is 10 m/s
    Fe = 500.      # sampling frequency of the system is 500 Hz
    tret = 0.004   # delay in the loop is 4 ms
    gain = 0.3     # closed-loop gain is 0.3

    # Here is the position of the object in the field
    offx, offy = (10., 0.)


    #################
    # PSF generation

    # Let's go. Let's define some basic parameters (arrays of spatial
    # frequencies)
    kx, ky, uk = computeSpatialFreqArrays(N, pixelSize, wavelengthIR)
    M4 = defineDmFrequencyArea(kx, ky, rotdegree)

    # This is the turbulent spectrum ....
    W = computeWiener(kx, ky, L0, r0IR)

    # And here are some of the PSF-destroyers - (English: Wavefront errors)
    Waniso = anisoplanaticSpectrum(Cn2h, layerAltitude, L0, offx, offy,
                                   wavelengthIR, kx, ky, W, M4)
    Wfit = fittingSpectrum(W, M4)
    Walias = aliasingSpectrum(kx, ky, r0IR, L0, M4)
    Wbp = computeBpSpectrum(kx, ky, V, Fe, tret, gain, W, M4)
    nmRms = 100.
    Wother = otherSpectrum(nmRms, M4, uk, wavelengthIR)

    # THE missing term = noise
    # Wnoise = noiseSpectrum(Rmag, .. kx, ky)  available one day ...

    # Now, you sum up every contributor, and produce a phase structure function
    Dphi = convertSpectrum2Dphi(Waniso + Wfit + Wother + Walias + Wbp, uk)

    # And you "blur" the nice Airy pattern using that phase structure function
    FTOtel = computeEeltOTF(pup)
    psf = core_generatePsf(Dphi, FTOtel)

    print('Strehl is ', psf.max())
    plt.imshow( np.log(psf) )

    return psf


#  _   _             ____                 _____
# | | | |___  ___   / ___|__ _ ___  ___  |___ /
# | | | / __|/ _ \ | |   / _` / __|/ _ \   |_ \
# | |_| \__ \  __/ | |__| (_| \__ \  __/  ___) |
#  \___/|___/\___|  \____\__,_|___/\___| |____/


def instantaneous_scao_psf():
    """

    The dirty one.
    Let's try to simulate the fluctuations due to short exposures.

    """
    # I start from "use case 2", and I sum the contributors to the phase error.
    # What I get is the total power spectrum of the perturbed phase.
    WW = Waniso + Wfit + Wother + Walias + Wbp

    # So, i'm gonna do some random draw of a phase that follows the statistics
    # of the spectrum WW. For that, i'm gonna use sqrt(WW) as the modulus of the
    # FFT of the phase, and generate a random phase chosen uniformly between 0
    # and 2.pi. I will then do a FFT of that, in order to get the phase.
    WW[0, 0] = 0        # because i don't care about piston mode
    WW = np.sqrt(WW)
    tmp = np.fft.fft2(WW * np.exp(2j * np.pi*np.random.rand(N, N))) * (uk)
    ph1 = tmp.real * np.sqrt(2)
    ph2 = tmp.imag * np.sqrt(2)

    # now i compute widthScreen, the size of the pixels of the phase screens I
    # have generated.
    widthScreen = 1. / uk  # in metres
    ud = widthScreen / N   # size of the pixels of the phase screen

    # With such a wide screen, and using a wind speed of V m/s, then I can
    # simulate an exposure time of  (widthScreen/V) seconds.
    # I recommend to sum psf snapshots every 50 cm (actuator pitch of M4).
    step = 0.50
    stepPix = int(np.round(step / ud))
    stepTime = (stepPix * ud) / V
    DIT = 1.0  # 1 second integration time
    niter = int(np.round(DIT / stepTime))
    psfLE = 0  # psf Long Exposure
    normFactor = np.sum(pup)**2
    for i in range(niter):
        psfSE = np.fft.fftshift(np.abs(np.fft.fft2(pup * np.exp(1j*ph1)))**2)
        psfSE /= normFactor
        print(psfSE.max())
        psfLE += psfSE
        ph1 = np.roll(ph1, stepPix, axis=0)
    psfLE /= niter

    # Here, possibilities are infinite ..
    # You can add some static aberrations, etc etc,
    # and generate all the PSFs you want.










import numpy as np
from astropy.io import fits
from astropy.table import Table

from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

from anisocado import AnalyticalScaoPsf


def strehl_map(r=25, dr=3, **kwargs):
    psf = AnalyticalScaoPsf(**kwargs)
    x, y = np.mgrid[-r:r+dr:dr, -r:dr+r:dr]

    strlmap = np.zeros(x.shape)
    for i in range(len(x)):
        for j in range(len(y)):
            psf.shift_off_axis(x[i, j], y[i, j])
            strlmap[i, j] = psf.strehl_ratio

    return strlmap


def on_axis_strehl_for_kernel_size(Narr=(128, 512, 2048), **kwargs):
    """Only for the on-axis kernel"""
    return [AnalyticalScaoPsf(N=N, **kwargs).strehl_ratio for N in Narr]


def make_psf_grid(r=14, dr=7, **kwargs):
    psf = AnalyticalScaoPsf(**kwargs)
    x, y = np.mgrid[-r:r+1:dr, -r:r+1:dr]

    psf_grid = []
    for i in range(len(x)):
        for j in range(len(y)):
            psf.shift_off_axis(x[i, j], y[i, j])
            psf_grid += [psf.kernel]

    return psf_grid


def make_image_of_psf_grid():
    psf_grid = make_psf_grid(wavelength=2.15, N=128)

    plt.figure(figsize=(10, 10))
    i = 0
    for y in range(5):
        for x in range(5):
            plt.subplot(5, 5, 1+x+5*(4-y))
            plt.imshow(psf_grid[i], origin="l", norm=LogNorm())
            plt.axis("off")
            plt.title("({}, {})".format((7*x-14), (7*x-14)))
            i += 1
    plt.suptitle("Ks-band (2.15um) SCAO PSFs")
    plt.show()


def make_simcado_psf_file(coords, wavelengths, header_cards=None, **kwargs):
    """
    Generate a set of Field-Varying PSF cubes for use with SimCADO

    Parameters
    ----------
    coords : list of tuples
        [arcsec] Sample positions of the PSF in field of view. (0, 0) is the
        centre of the field of view.

    wavelengths : list
        [um] The wavelengths for which the PSF should be sampled

    header_cards : dict, optional
        Any extra keyword-value pair to be added to the extension 0 header.

    kwargs
    ------
    Keyword-value pairs accepted by an ``AnalyticalScaoPsf`` object

    Returns
    -------
    hdulist : fits.HDUList
        A HDUList object which is formatted for use as a Field-Varying PSF in
        SimCADO

    Examples
    --------
    ::

        import anisocado

        radii = [1, 3, 5, 10]   # arcsec
        waves = [1.0, 1.2, 1.6, 2.15]   # um

        coords = anisocado.field_positions_for_simcado_psf(radii, theta=60)
        hdu = anisocado.make_simcado_psf_file(coords=coords, wavelengths=waves,
                                              N=512, profile_name="EsoQ1")

        hdu.writeto("new_scao_psf_cube.fits")

    """

    # accept list of coordinates
    # accept list of wavelengths
    # make ext0
    # make table for ext1
    # initialise PSF
    # shift to each positions
    # make hdus for each position
    # build hdulist
    # return hdulist

    x, y = np.array(coords).T
    layers = np.arange(len(x))

    keys = ["AUTHOR", "DATE_CRE", "DATE_MOD", "SOURCE", "STATUS"]
    ext0_dict = {key: "" for key in keys}
    ext0_dict["ETYPE"] = "FVPSF"
    ext0_dict["ECAT"] = (1, "The extension containing the catalogue data")
    ext0_dict["EDATA"] = (2, "The first extension with real data")
    ext0_dict.update({"WAVEEXT{}".format(i + 2): w
                      for i, w in enumerate(wavelengths)})

    pri_hdr = fits.PrimaryHDU()
    pri_hdr.header.update(ext0_dict)
    pri_hdr.header.update(header_cards)

    ext1_dict = {"NUMPSFS": len(x), "CATTYPE": "table", "CUNIT1": "arcsec"}

    tbl = Table(data=[x, y, layers], names=["x", "y", "layer"])
    cat_hdu = fits.table_to_hdu(tbl)
    cat_hdu.header.update(ext1_dict)

    psf_hdus = []
    for wave in wavelengths:
        print("Making psf cube for {} um".format(wave))
        psf = AnalyticalScaoPsf(wavelength=wave, **kwargs)
        kernel_cube = [psf.shift_off_axis(dx, dy) for dx, dy in coords]

        psf_hdu = psf.hdu
        psf_hdu.data = np.array(kernel_cube)
        psf_hdu.header["WAVE0"] = wave
        psf_hdu.header["WAVEUNIT"] = "um"

        psf_hdus += [psf_hdu]

    hdulist = fits.HDUList([pri_hdr, cat_hdu] + psf_hdus)

    return hdulist


def field_positions_for_simcado_psf(radii=None, theta=45):
    """
    Generates a list of field position where the PSF will be sampled

    The PSF will be sampled at intervals of ``theta`` around concentric circles
    placed at distances ``radii`` from the centre of the field of view.

    Default radii are at [1, 2, 4, 8, 16, 32] arcsec

    Parameters
    ----------
    radii : list of floats
        [arcsec] Radii of concentric circles where the PSF will be sampled

    theta : float
        [deg] Spacing between PSF samples around each circle

    Returns
    -------
    coords : list of tuples
        [arcsec] List of sample positions relative to the centre of the field

    Examples
    --------
    ::

        import anisocado
        cds = anisocado.field_positions_for_simcado_psf(radii=[5, 10], theta=60)
        hdu = anisocado.make_simcado_psf_file(cds, wavelengths=[1.2, 1.6, 2.2])

    """

    coords = [(0, 0)]
    if radii is None:
        radii =  [1, 2, 4, 8, 16, 32]
    for r in radii:
        for ang in np.arange(0, 360, theta):
            coords += [(r * np.cos(np.deg2rad(ang)),
                        r * np.sin(np.deg2rad(ang)))]

    return coords


def make_strehl_map_from_coords(coords):
    x, y = np.array(coords).T

    from scipy.interpolate import griddata
    map = griddata((x, y), np.arange(len(x)),
                   np.array(np.meshgrid(np.arange(-25, 26),
                                        np.arange(-25, 26))).T,
                   method="nearest")

    return map





# -*- coding: utf-8 -*-
# Learn more: https://github.com/kennethreitz/setup.py

from setuptools import setup, find_packages
from io import open 

with open('README.md', "r", encoding='utf-8') as f:
    __readme__ = f.read()

with open('LICENSE') as f:
    __license__ = f.read()

__version__ = "0.1.rc1"

setup(
    name='anisocado',
    version=__version__,
    description='Generate off-axis SCAO PSFs for MICADO at the ELT',
     author='Eric Gendron, Kieran Leschinski',
    author_email='kieran.leschinski@univie.ac.at',
    url='https://github.com/astronomyk/anisocado',
    license="GNU GENERAL PUBLIC LICENSE Version 2, June 1991",
    include_package_data=True,
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=['numpy', 'astropy', 'matplotlib'],
    long_description=__readme__,
    long_description_content_type = 'text/markdown'
   )



import warnings
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

from astropy.io import fits

from .psf_utils import *


class AnalyticalScaoPsf:
    """
    A class to generate SCAO PSFs for MICADO at the ELT

    It is important to note that original PSF is generated for an on-axis
    guide star (``self.psf_on_axis``) at a specific wavelength. This PSF is
    kept as a reference for all further operations.

    When the guide star is shifted off axis, the new PSF kernel is returned by
    the function ``self.shift_off_axis()``. The kernel is also kept in
    ``self.psf_latest``.

    The property ``self.strehl_ratio`` is always calculated from
     ``self.psf_latest`` (or ``self.psf_on_axis`` if no shift has been applied).

    Parameters
    ----------
    N : int
        [pixel] Default: 512 pixel. Side-length of the kernel array
    pixelSize : float
        [arcsec] Default: 0.004 arcsec. On-sky pixel scale
    wavelength : float
        [um] Default: 2.15 um. Wavelength for which the PSF should be generated
    rotdegree : float
        [deg] Default: 0 deg. Rotation angle of the pupil w.r.t the plane of the optical axis
    nmRms : float
        [nm] Default: 100 nm. Residual wavefront error of the system
    L0 : float
        [m] Default: 25 m. Outer scale
    profile_name : str
        ['EsoQ1', 'EsoMedian', 'EsoQ4', 'oldEso', 'gendron']. Default: EsoMedian
        Names of specific atmospheric conditions for which presets exist.
        See ``psf_utils.get_atmospheric_turbulence``
    deadSegments : int
        Default: 5. Number of segments missing from the ELT primary mirror
    V : float
        [m/s] Default: 10 m/s. Wind speed
    Fe : float
        [Hz] Default: 500 Hz. AO sampling frequency of the system
    tret : float
        [s] Default: 0.004 s. Delay in the AO loop
    gain : float
        Default: 0.3. Closed-loop gain
    dactu : float
        [m] Default: 0.5403 m. Interactuator distance on M4
    x_last, y_last : float
        [arcsec] Default: 0 arcsec. Shifts used to generate the ``psf_latest``
        kernel

    Derived Attributes
    ------------------
    seeing : float
        [arcsec] Default: 0.67 arcsec. Set by profile_name, if not set by user
    zenDist : float
        [degree] Default: 30 deg. Zenith distance. Set by profile_name, if not
        set by user
    r0Vis : float
        [m] Fried parameter at 500 nm.
    r0IR : float
        [m] Fried parameter at 500 nm.
    psf_on_axis : array
        The on-axis PSF kernel
    psf_latest : array
        The last PSF generated or shifted is kept in memory
    pup : array
        An image of the telescope pupil, i.e. ELT M1
    layerAltitude: list, array
        [m] The heights of the different turbulent layers
    Cn2h : array, list
        [0..1] The relative strength of each turbulent layer
    kx, ky : 2D meshgrid arrays
        [m-1] spatial frequencies - see ``utils.computeSpatialFreqArrays``
    uk : array
        [mas / m]
    M4 : array
        2D domain of spatial frequencies which can be impacted by M4, i.e. the
        M4 compensation domain. See ``utils.defineDmFrequencyArea``
    W :
        [rad2 m2] 2D spectrum of the turbulence with an outer scale L0.
        See ``utils.computeWiener``

    Examples
    --------
    Make a PSF for a NGS 5 arcsec off axis and save it to disk::

        from anisocado import AnalyticalScaoPsf
        psf = AnalyticalScaoPsf(N=512, wavelength=2.15, seeing=0.8)
        psf.shift_off_axis(5, 0)
        psf.writeto("my_psf.fits")

    """

    def __init__(self, **psf_on_axis):
        self.kwarg_names = ["N", "pixelSize", "wavelength", "rotdegree",
                            "seeing", "r0Vis", "r0IR", "nmRms", "L0",
                            "profile_name", "zenDist", "deadSegments",
                            "V", "Fe", "tret", "gain", "dactu", "x_last",
                            "y_last"]

        # Variable attributes
        self.N = 512
        self.pixelSize = 0.004      # arcsec
        self.wavelength = 2.15      # um
        self.rotdegree = 0.         # deg
        self.nmRms = 100.           # nm
        self.L0 = 25.               # meters - from doc. ESO-258292.
        self.profile_name = "EsoMedian"  # "oldEso", "EsoMedian, EsoQ1, EsoQ4"
        self.deadSegments = 5       # there are some missing segments tonight !
        self.V = 10.                # wind is 10 m/s
        self.Fe = 500.              # sampling frequency of the system is 500 Hz
        self.tret = 0.004           # [s] delay in the loop is 4 ms
        self.gain = 0.3             # closed-loop gain is 0.3
        self.dactu = 0.5403         # [m] distance betweem actuators on M4
        self.x_last = 0             # [arcsec] x shift used to make psf_latest
        self.y_last = 0             # [arcsec] x shift used to make psf_latest

        # Derived attributes
        self.seeing = None          # arcsec - set by profile_name if None
        self.zenDist = None         # degree - set by profile_name if None
        self.r0Vis = None           # meters
        self.r0IR = None            # meters
        self.psf_on_axis = None
        self.psf_latest = None
        self.pup = None
        self.layerAltitude = None
        self.Cn2h = None
        self.kx = None
        self.ky = None
        self.uk = None
        self.M4 = None
        self.W = None
        self._wave_m = None
        self._kernel_sum = None
        
        self.__dict__.update(psf_on_axis)
        self.update()

    def update(self, **kwargs):
        """
        Updates the parameter needed to generate a PSF and/or shift if off-axis

        Valid parameter names can be found in ``self.kwarg_names``

        """
        self._wave_m = self.wavelength
        if self.wavelength > 0.1:  # assume its in um
            self._wave_m *= 1e-6
        
        for key in kwargs:
            if key not in self.kwarg_names:
                warnings.warn("{} not found in self.kwarg_names".format(key))

        self.__dict__.update(kwargs)

        # and layers appear further away with zenith distance
        if self.profile_name is not None:
            seeing, zen_dist, wind_a = get_profile_defaults(self.profile_name)
            self.V *= wind_a
            if self.seeing is None:
                self.seeing = seeing
            if self.zenDist is None:
                self.zenDist = zen_dist

            layerAltitude, Cn2h = get_atmospheric_turbulence(self.profile_name)
            self.layerAltitude = np.array(layerAltitude, dtype=float)
            self.Cn2h = Cn2h

        self.layerAltitude *= 1 / np.cos(self.zenDist * np.pi / 180.)

        # r0Vis is in metres here (0.103 is in metres.arcsec)
        # convert r0 from 500nm to IR
        # apparent seeing degrades with airmass
        self.r0Vis = 0.103 / self.seeing
        self.r0IR = r0Converter(self.r0Vis, 500e-9, self._wave_m)
        self.r0IR = airmassImpact(self.r0IR, self.zenDist)

        # generate the ELT pupil
        self.pup = fake_generatePupil(self.N, self.deadSegments, self.rotdegree,
                                      self.pixelSize, self._wave_m)

        # Let's go. Let's define some basic parameters
        # (arrays of spatial frequencies)
        self.kx, self.ky, self.uk = computeSpatialFreqArrays(self.N,
                                                             self.pixelSize,
                                                             self._wave_m)
        self.M4 = defineDmFrequencyArea(self.kx, self.ky, self.rotdegree,
                                        self.dactu)

        # This is the turbulent spectrum ....
        self.W = computeWiener(self.kx, self.ky, self.L0, self.r0IR)

        if self.psf_on_axis is None:
            self.psf_on_axis = self.make_psf()
            self.psf_latest = self.psf_on_axis

    def make_psf(self):
        """
        Generates a analytical SCAO PSF for a long (>10 sec) exposure

        Parameters need to be set be setting the attibute directly, or by
        calling ``self.update()`` with the desired keyword-value pair passed as
         a kwarg. Valid keywords can be found in ``self.kwarg_names``.

        Returns
        -------
        psf : array
            The PSF kernel

        Examples
        --------


        """
        # Initial setup happens in update()

        dx, dy = 0, 0
        # And here are some of the PSF-destroyers  (Translation: Wavefront errs)
        Waniso = anisoplanaticSpectrum(self.Cn2h, self.layerAltitude, self.L0,
                                       dx, dy, self._wave_m,
                                       self.kx, self.ky, self.W, self.M4)
        Wfit = fittingSpectrum(self.W, self.M4)
        Walias = aliasingSpectrum(self.kx, self.ky, self.r0IR, self.L0, self.M4)
        Wbp = computeBpSpectrum(self.kx, self.ky, self.V, self.Fe, self.tret,
                                self.gain, self.W, self.M4)
        Wother = otherSpectrum(self.nmRms, self.M4, self.uk, self._wave_m)
        # Wnoise = noiseSpectrum(Rmag, .. kx, ky)  available one day ...

        # Now, you sum up every contributor, and produce a phase structure funcn
        Wtotal = Waniso + Wfit + Wother + Walias + Wbp  # + Wnoise
        # Wtotal = np.zeros(np.shape(Wtotal))
        Dphi = convertSpectrum2Dphi(Wtotal, self.uk)

        # And you "blur" the nice Airy pattern using that phase structure funcn
        FTOtel = computeEeltOTF(self.pup)
        psf = core_generatePsf(Dphi, FTOtel)
        sum_kernel = np.sum(psf)
        if self._kernel_sum is None:
            self._kernel_sum = sum_kernel
        psf /= sum_kernel
        # psf = clean_psf(psf, 1E-7)
        self.psf_latest = psf

        return psf

    def shift_off_axis(self, dx, dy):
        """
        Shifts the on-axis PSF off axis by an amount ``(dx, dy)`` in arcsec

        Parameters
        ----------
        dx, dy : float
            [arcsec] Offset in each of the dimensions relative to the plane of
            the optical axis

        Returns
        -------
        psf : array
            The PSF kernel

        Examples
        --------


        """

        self.x_last = dx
        self.y_last = dy

        # Original setup starts in update()

        # and all this will be used to run the function below, that will
        # compute the spatial spectrum of the phase due to anisoplanatism
        Waniso = anisoplanaticSpectrum(self.Cn2h, self.layerAltitude, self.L0,
                                       dx, dy, self._wave_m,
                                       self.kx, self.ky, self.W, self.M4)

        # Transforming this spectrum into a phase structure function
        Dphi = convertSpectrum2Dphi(Waniso, self.uk)

        # Here, the on-axis psf comes into play ... I take its Fourier transform
        # it's complex.
        fto = np.fft.fft2(np.fft.fftshift(self.psf_on_axis)) / self.N ** 2

        psf = core_generatePsf(Dphi, fto)
        psf /= np.sum(psf)
        self.psf_latest = psf

        return psf

    def make_short_exposure_psf(self, dit=1.0, screen_step_length=0.5):
        """
        Returns a PSF for an 'short' exposure time of ``DIT``

        The PSF kernel will be a single 2D array made from N stacked
        instantaneous PSFs, where the instantaneous PSFs are generated at time
        intervals during the DIT length determined by the wind speed,
        ``self.V``, and the phase-screen ``step`` length.

        Parameters
        ----------
        dit : float
            [s] Default is 1.0 sec. Exposure time for the PSF

        screen_step_length : float
            [m] Sample step length for atmospheric phase screen
            Default is 0.5m - the length of the M4 actuator pitch

        Returns
        -------
        psfLE : array

        Examples
        --------


        """

        # The dirty one.
        # Let's try to simulate the fluctuations due to short exposures.
        Waniso = anisoplanaticSpectrum(self.Cn2h, self.layerAltitude, self.L0,
                                       self.x_last, self.x_last,
                                       self._wave_m, self.kx, self.ky,
                                       self.W, self.M4)
        Wfit = fittingSpectrum(self.W, self.M4)
        Walias = aliasingSpectrum(self.kx, self.ky, self.r0IR, self.L0, self.M4)
        Wbp = computeBpSpectrum(self.kx, self.ky, self.V, self.Fe, self.tret,
                                self.gain, self.W, self.M4)
        Wother = otherSpectrum(self.nmRms, self.M4, self.uk, self._wave_m)
        # Wnoise = noiseSpectrum(Rmag, .. kx, ky)  available one day ...

        # I start from "use case 2", and I sum the contributors to the phase err
        # What I get is the total power spectrum of the perturbed phase.
        Wtotal = Waniso + Wfit + Wother + Walias + Wbp

        # So, i'm gonna do some random draw of a phase that follows the
        # statistics of the spectrum WW. For that, i'm gonna use sqrt(WW) as
        # the modulus of the FFT of the phase, and generate a random phase
        # chosen uniformly between 0 and 2.pi. I will then do a FFT of that, in
        # order to get the phase.
        Wtotal[0, 0] = 0  # because i don't care about piston mode
        Wtotal = np.sqrt(Wtotal)
        rand_phase = np.random.rand(self.N, self.N)
        tmp = np.fft.fft2(Wtotal * np.exp(2j * np.pi * rand_phase)) * self.uk
        ph1 = tmp.real * np.sqrt(2)
        ph2 = tmp.imag * np.sqrt(2)

        # now i compute widthScreen, the size of the pixels of the phase screens
        # I have generated.
        widthScreen = 1. / self.uk  # in metres
        ud = widthScreen / self.N  # size of the pixels of the phase screen

        # With such a wide screen, and using a wind speed of V m/s, then I can
        # simulate an exposure time of  (widthScreen/V) seconds.
        stepPix = int(np.round(screen_step_length / ud))
        stepTime = (stepPix * ud) / self.V
        niter = int(np.round(dit / stepTime))
        psfLE = 0  # psf Long Exposure
        normFactor = np.sum(self.pup) ** 2
        for i in range(niter):
            inv_pupil = np.fft.fft2(self.pup * np.exp(1j * ph1))
            psfSE = np.fft.fftshift(np.abs(inv_pupil) ** 2)
            psfSE /= normFactor
            # print(psfSE.max())
            psfLE += psfSE
            ph1 = np.roll(ph1, stepPix, axis=0)
        psfLE /= niter
        psfLE /= np.sum(psfLE)
        self.psf_latest = psfLE

        return psfLE

    @property
    def strehl_ratio(self):
        """Return an Strehl ratio of the kernel in ``self.psf_latest``"""
        return np.max(self.psf_latest) * self._kernel_sum

    @property
    def kernel(self):
        """Return the kernel held in ``self.psf_latest``"""
        return self.psf_latest

    @property
    def hdu(self):
        """Return the ``ImageHDU`` from ``get_hdu()``"""
        return self.get_hdu()

    def writeto(self, **kwargs):
        """Calls the ``writeto`` method of the ImageHDU from ``self.hdu``"""
        self.hdu.writeto(**kwargs)

    def get_hdu(self, **kwargs):
        """
        Makes an ``ImageHDU`` with the kernel and relevant header info

        Additional keyword-value pairs can be passed to the header as kwargs

        Returns
        -------
        hdu_psf : fits.ImageHDU

        """
        w, h = self.psf_latest.shape

        hdr = fits.Header()
        hdr["CDELT1"] = self.pixelSize / 3600.  # because pixelSize is in arcsec
        hdr["CDELT2"] = self.pixelSize / 3600.  
        hdr["CRVAL1"] = self.x_last / 3600.
        hdr["CRVAL2"] = self.y_last / 3600.
        hdr["CRPIX1"] = w / 2.
        hdr["CRPIX2"] = h / 2.
        hdr["CTYPE1"] = "RA---TAN"
        hdr["CTYPE2"] = "DEC--TAN"
        hdr["CUNIT1"] = "degree"
        hdr["CUNIT2"] = "degree"
        hdr["WAVE0"] = (self.wavelength, "[um] Central wavelength")
        hdr["WAVEUNIT"] = "um"
        hdr["STREHL"] = (self.strehl_ratio, "Strehl ratio")
        hdr["PSFSUM"] = self._kernel_sum, "Sum of on-axis kernel used for SR"

        dic = {key: self.__dict__[key] for key in self.kwarg_names}
        hdr.update(dic)
        hdr.update(kwargs)
        hdu_psf = fits.ImageHDU(data=self.psf_latest, header=hdr)

        return hdu_psf

    def plot_psf(self, which="psf_latest"):
        """Plots a logscale PSF kernel: ["psf_latest", "psf_on_axis"]"""
        plt.imshow(getattr(self, which).T, origin='l', norm=LogNorm())
        print('Strehl ratio of {} is {}'.format(which, self.psf_latest.max()))


